#!/usr/bin/env python3
"""
Script to update Colab links in notebooks to match the current git branch.
This ensures that when documentation is built from different branches,
the Colab links point to the correct branch.
"""
import json
import os
import subprocess
import sys
from pathlib import Path


def get_git_branch():
    """Get the current git branch name."""
    # First, try to get from environment variable (set by CI/CD)
    branch = os.environ.get('GIT_BRANCH')
    
    if branch:
        return branch
    
    # If not in CI, try to detect from git
    try:
        result = subprocess.run(
            ['git', 'rev-parse', '--abbrev-ref', 'HEAD'],
            capture_output=True,
            text=True,
            check=True
        )
        return result.stdout.strip()
    except (subprocess.CalledProcessError, FileNotFoundError):
        # Fallback to 'develop' if git is not available
        return 'develop'


def update_notebook_links(notebook_path: Path, target_branch: str):
    """Update Colab links in a notebook to point to the target branch."""
    try:
        with open(notebook_path, 'r', encoding='utf-8') as f:
            notebook = json.load(f)
        
        modified = False
        
        # Define replacement patterns
        github_base = 'https://github.com/Mitchell-Mirano/sorix'
        github_blob = f'{github_base}/blob/{target_branch}'
        colab_base = f'https://colab.research.google.com/github/Mitchell-Mirano/sorix/blob/{target_branch}'
        
        # Documentation base URL - use GitHub Pages in CI, localhost in local dev
        is_ci = os.environ.get('CI', '').lower() == 'true'
        if is_ci:
            docs_base = 'https://mitchell-mirano.github.io/sorix'
        else:
            docs_base = 'http://127.0.0.1:8000'
        
        # Process all cells
        for cell in notebook.get('cells', []):
            cell_type = cell.get('cell_type')
            if cell_type in ['markdown', 'code']:
                source = cell.get('source', [])
                if isinstance(source, list):
                    new_source = []
                    for line in source:
                        new_line = line
                        
                        # Update pip install command (usually in code cells)
                        if 'pip install' in new_line and 'sorix @ git+' in new_line:
                            import re
                            replaced_line = re.sub(
                                r'(github\.com/Mitchell-Mirano/sorix\.git@)[^/\'"]+',
                                rf'\1{target_branch}',
                                new_line
                            )
                            if replaced_line != new_line:
                                new_line = replaced_line
                                modified = True

                        # Replace Jinja2 macros with actual values
                        if '{{ github_blob }}' in new_line:
                            new_line = new_line.replace('{{ github_blob }}', github_blob)
                            modified = True
                        
                        if '{{ colab_base }}' in new_line:
                            new_line = new_line.replace('{{ colab_base }}', colab_base)
                            modified = True
                        
                        if '{{ github_base }}' in new_line:
                            new_line = new_line.replace('{{ github_base }}', github_base)
                            modified = True
                        
                        if '{{ branch }}' in new_line:
                            new_line = new_line.replace('{{ branch }}', target_branch)
                            modified = True
                        
                        if '{{ docs_base }}' in new_line:
                            new_line = new_line.replace('{{ docs_base }}', docs_base)
                            modified = True
                        
                        # Also replace hardcoded Colab links (for backward compatibility)
                        if cell_type == 'markdown' and 'colab.research.google.com/github/Mitchell-Mirano/sorix/blob/' in new_line:
                            import re
                            replaced_line = re.sub(
                                r'(colab\.research\.google\.com/github/Mitchell-Mirano/sorix/blob/)[^/]+/',
                                rf'\1{target_branch}/',
                                new_line
                            )
                            if replaced_line != new_line:
                                new_line = replaced_line
                                modified = True
                        
                        # Replace hardcoded GitHub blob links
                        if cell_type == 'markdown' and 'github.com/Mitchell-Mirano/sorix/blob/' in new_line and 'colab' not in new_line:
                            import re
                            replaced_line = re.sub(
                                r'(github\.com/Mitchell-Mirano/sorix/blob/)[^/]+/',
                                rf'\1{target_branch}/',
                                new_line
                            )
                            if replaced_line != new_line:
                                new_line = replaced_line
                                modified = True
                        
                        new_source.append(new_line)
                    
                    cell['source'] = new_source
        
        # Write back if modified
        if modified:
            with open(notebook_path, 'w', encoding='utf-8') as f:
                json.dump(notebook, f, indent=1, ensure_ascii=False)
            try:
                rel_path = notebook_path.relative_to(Path.cwd())
            except ValueError:
                rel_path = notebook_path
            print(f"✓ Updated {rel_path}")
            return True
        
        return False
        
    except Exception as e:
        print(f"✗ Error processing {notebook_path}: {e}", file=sys.stderr)
        return False


def main():
    """Main function to update all notebooks."""
    # Get the target branch
    branch = get_git_branch()
    print(f"Updating notebook links to branch: {branch}")
    
    # Find all notebooks in docs directory
    docs_dir = Path('docs')
    if not docs_dir.exists():
        print(f"Error: docs directory not found", file=sys.stderr)
        sys.exit(1)
    
    notebooks = list(docs_dir.rglob('*.ipynb'))
    
    if not notebooks:
        print("No notebooks found")
        return
    
    print(f"Found {len(notebooks)} notebook(s)")
    
    updated_count = 0
    for notebook in notebooks:
        if update_notebook_links(notebook, branch):
            updated_count += 1
    
    print(f"\nUpdated {updated_count}/{len(notebooks)} notebook(s)")


if __name__ == '__main__':
    main()



import pytest
import numpy as np
from sorix import Tensor, tensor
from sorix.utils import utils

def test_sigmoid():
    x = tensor([0.0], requires_grad=True)
    y = utils.sigmoid(x)
    assert y.data == 0.5
    y.backward()
    assert x.grad == 0.25 # 0.5 * (1 - 0.5)
    
    # Non-tensor
    assert utils.sigmoid(0.0) == 0.5

def test_softmax():
    x = tensor([[1.0, 2.0, 3.0]], requires_grad=True)
    y = utils.softmax(x)
    expected = np.exp(np.array([[1.0, 2.0, 3.0]]))
    expected /= expected.sum()
    assert np.allclose(y.data, expected)
    
    # Non-tensor
    assert np.allclose(utils.softmax(np.array([[1.0, 2.0, 3.0]])), expected)

def test_argmax_argmin():
    x = tensor([[1.0, 5.0, 2.0], [4.0, 0.0, 6.0]])
    
    am = utils.argmax(x, axis=1)
    assert np.array_equal(am.data, [[1], [2]])
    
    an = utils.argmin(x, axis=1)
    assert np.array_equal(an.data, [[0], [1]])
    
    # Non-tensor
    xn = np.array([[1.0, 5.0, 2.0], [4.0, 0.0, 6.0]])
    assert np.array_equal(utils.argmax(xn, axis=1), [[1], [2]])
    assert np.array_equal(utils.argmin(xn, axis=1), [[0], [1]])

def test_as_tensor_from_numpy():
    data = [1, 2, 3]
    t1 = utils.as_tensor(data)
    assert isinstance(t1, Tensor)
    assert np.array_equal(t1.data, data)
    
    t2 = utils.as_tensor(t1)
    assert t2 is t1
    
    t3 = utils.from_numpy(np.array(data))
    assert isinstance(t3, Tensor)
    assert np.array_equal(t3.data, data)
    
    t4 = utils.from_numpy(t3)
    assert t4 is t3

def test_creation_ops():
    # zeros
    z = utils.zeros((2, 3))
    assert z.shape == (2, 3)
    assert np.all(z.data == 0)
    
    # ones
    o = utils.ones((2, 2))
    assert o.shape == (2, 2)
    assert np.all(o.data == 1)
    
    # full
    f = utils.full((2, 2), 7)
    assert np.all(f.data == 7)
    
    # eye
    e = utils.eye(3)
    assert np.array_equal(e.data, np.eye(3))
    
    # diag
    d = utils.diag([1, 2, 3])
    assert np.array_equal(d.data, np.diag([1, 2, 3]))
    
    # arange
    ar = utils.arange(0, 10, 2)
    assert np.array_equal(ar.data, np.arange(0, 10, 2))
    
    # linspace
    ls = utils.linspace(0, 1, 5)
    assert np.array_equal(ls.data, np.linspace(0, 1, 5))
    
    # logspace
    lgs = utils.logspace(0, 2, 3)
    assert np.array_equal(lgs.data, np.logspace(0, 2, 3))

def test_random_ops():
    # rand
    r = utils.rand(10, 10)
    assert r.shape == (10, 10)
    assert np.all(r.data >= 0) and np.all(r.data <= 1)
    
    # randn
    rn = utils.randn(10, 10)
    assert rn.shape == (10, 10)
    
    # randint
    ri = utils.randint(0, 10, (5, 5))
    assert ri.shape == (5, 5)
    assert np.all(ri.data >= 0) and np.all(ri.data < 10)
    
    # randperm
    rp = utils.randperm(10)
    assert len(rp.data) == 10
    assert sorted(rp.data.tolist()) == list(range(10))

def test_like_ops():
    x = tensor([[1, 2], [3, 4]])
    
    zl = utils.zeros_like(x.data)
    assert zl.shape == (2, 2)
    assert np.all(zl.data == 0)
    
    ol = utils.ones_like(x.data)
    assert ol.shape == (2, 2)
    assert np.all(ol.data == 1)
    
    fl = utils.full_like(x.data, 5)
    assert np.all(fl.data == 5)
    
    el = utils.empty_like(x.data)
    assert el.shape == (2, 2)

def test_gpu_logic_if_available():
    from sorix.cupy.cupy import _cupy_available
    if _cupy_available:
        # Test creation on GPU
        z = utils.zeros((2, 2), device='gpu')
        assert z.device == 'gpu'
        import cupy as cp
        assert isinstance(z.data, cp.ndarray)
        
        # Test argmax on GPU
        x = tensor([[1, 5, 2]], device='gpu')
        am = utils.argmax(x)
        assert am.device == 'gpu'
        assert am.data[0, 0] == 1
        
        # Test softmax on GPU
        sm = utils.softmax(x)
        assert sm.device == 'gpu'
        
        # Test ones_like on GPU
        ol = utils.ones_like(x.data, device='gpu')
        assert ol.device == 'gpu'
        assert isinstance(ol.data, cp.ndarray)



import pytest
from sorix.cuda import cuda
from sorix.cupy.cupy import _cupy_available

def test_cuda_availability_check():
    """Verify that the CUDA availability check runs without errors."""
    available = cuda.is_available(verbose=False)
    assert isinstance(available, bool)
    
    # If copper is not available, it MUST be false
    if not _cupy_available:
        assert available == False

def test_cupy_flag_exists():
    assert isinstance(_cupy_available, bool)



import numpy as np

class DataLoader:
    def __init__(self, dataset, batch_size=16, shuffle=True):
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.indices = np.arange(len(dataset))

    def __iter__(self):
        if self.shuffle:
            np.random.shuffle(self.indices)

        for i in range(0, len(self.indices), self.batch_size):
            batch_indices = self.indices[i : i + self.batch_size]
            yield self.dataset[batch_indices]

    def __len__(self):
        return (len(self.dataset) + self.batch_size - 1) // self.batch_size

    


import pytest
import numpy as np
from sorix import Tensor, tensor
from sorix.utils import math

def test_add():
    a = tensor([1.0, 2.0], requires_grad=True)
    b = tensor([3.0, 4.0], requires_grad=True)
    
    # Test with tensor
    c = math.add(a, b)
    assert isinstance(c, Tensor)
    assert np.array_equal(c.data, [4.0, 6.0])
    
    c.backward()
    assert np.array_equal(a.grad, [1.0, 1.0])
    
    # Test with non-tensor
    assert math.add(1, 2) == 3

def test_sub():
    a = tensor([5.0], requires_grad=True)
    b = tensor([2.0], requires_grad=True)
    
    c = math.sub(a, b)
    assert np.array_equal(c.data, [3.0])
    c.backward()
    assert a.grad == [1.0]
    assert b.grad == [-1.0]
    
    assert math.sub(10, 3) == 7

def test_mul():
    a = tensor([2.0], requires_grad=True)
    b = tensor([4.0], requires_grad=True)
    
    c = math.mul(a, b)
    assert np.array_equal(c.data, [8.0])
    c.backward()
    assert a.grad == [4.0]
    assert b.grad == [2.0]
    
    assert math.mul(3, 4) == 12

def test_div():
    a = tensor([10.0], requires_grad=True)
    b = tensor([2.0], requires_grad=True)
    
    c = math.div(a, b)
    assert np.array_equal(c.data, [5.0])
    c.backward()
    assert a.grad == [1/2.0]
    assert b.grad == [-10.0/4.0]
    
    assert math.div(20, 5) == 4

def test_matmul():
    a = tensor([[1.0, 2.0]], requires_grad=True)
    b = tensor([[3.0], [4.0]], requires_grad=True)
    
    c = math.matmul(a, b)
    assert np.array_equal(c.data, [[11.0]])
    c.backward()
    # grad_a = grad_out @ b.T = [[1]] @ [[3, 4]] = [[3, 4]]
    assert np.array_equal(a.grad, [[3.0, 4.0]])
    
    # Test with non-tensor
    x = np.array([[1, 2]])
    y = np.array([[3], [4]])
    assert np.array_equal(math.matmul(x, y), [[11]])

def test_pow():
    a = tensor([3.0], requires_grad=True)
    c = math.pow(a, 2)
    assert np.array_equal(c.data, [9.0])
    c.backward()
    assert a.grad == [6.0]
    
    assert math.pow(2, 3) == 8

def test_sin():
    x = tensor([0.0, np.pi/2], requires_grad=True)
    y = math.sin(x)
    assert np.allclose(y.data, [0.0, 1.0])
    
    y.backward()
    assert np.allclose(x.grad, [1.0, 0.0]) # cos(0)=1, cos(pi/2)=0
    
    # Non-tensor
    assert np.allclose(math.sin(np.array([0.0])), [0.0])

def test_cos():
    x = tensor([0.0, np.pi/2], requires_grad=True)
    y = math.cos(x)
    assert np.allclose(y.data, [1.0, 0.0])
    
    y.backward()
    assert np.allclose(x.grad, [0.0, -1.0]) # -sin(0)=0, -sin(pi/2)=-1
    
    # Non-tensor
    assert np.allclose(math.cos(np.array([0.0])), [1.0])

def test_tanh():
    x = tensor([0.0], requires_grad=True)
    y = math.tanh(x)
    assert y.data == 0.0
    y.backward()
    assert x.grad == 1.0 # 1 - tanh(0)^2 = 1
    
    # Non-tensor
    assert math.tanh(0.0) == 0.0

def test_exp():
    x = tensor([0.0, 1.0], requires_grad=True)
    y = math.exp(x)
    assert np.allclose(y.data, [1.0, np.exp(1.0)])
    
    y.backward()
    assert np.allclose(x.grad, [1.0, np.exp(1.0)])
    
    # Non-tensor
    assert np.allclose(math.exp(np.array([0.0])), [1.0])

def test_log():
    x = tensor([1.0, np.exp(1.0)], requires_grad=True)
    y = math.log(x)
    assert np.allclose(y.data, [0.0, 1.0])
    
    y.backward()
    assert np.allclose(x.grad, [1.0, 1.0/np.exp(1.0)])
    
    # Non-tensor
    assert np.allclose(math.log(np.array([np.exp(1.0)])), [1.0])

def test_sqrt():
    x = tensor([4.0, 9.0], requires_grad=True)
    y = math.sqrt(x)
    assert np.allclose(y.data, [2.0, 3.0])
    
    y.backward()
    assert np.allclose(x.grad, [1/(2*2.0), 1/(2*3.0)])
    
    # Non-tensor
    assert math.sqrt(4.0) == 2.0

def test_mean():
    x = tensor([[1.0, 2.0], [3.0, 4.0]], requires_grad=True)
    y = math.mean(x)
    assert y.data == 2.5
    y.backward()
    assert np.allclose(x.grad, 0.25)
    
    # Non-tensor
    assert math.mean(np.array([1, 2, 3])) == 2.0

def test_sum():
    x = tensor([[1.0, 2.0], [3.0, 4.0]], requires_grad=True)
    y = math.sum(x)
    assert y.data == 10.0
    y.backward()
    assert np.array_equal(x.grad, [[1.0, 1.0], [1.0, 1.0]])
    
def test_gpu_math_if_available():
    from sorix.cupy.cupy import _cupy_available
    if _cupy_available:
        import cupy as cp
        x = tensor([0.0, np.pi/2], requires_grad=True, device='gpu')
        y = math.sin(x)
        assert y.device == 'gpu'
        assert isinstance(y.data, cp.ndarray)
        assert np.allclose(y.to_numpy(), [0.0, 1.0])
        y.backward()
        assert x.device == 'gpu'
        
        # Test mean/sum on GPU
        assert math.mean(x).device == 'gpu'
        assert math.sum(x).device == 'gpu'



import pytest
import numpy as np
from sorix import Tensor, tensor
from sorix.nn import Linear, ReLU, Sigmoid, Tanh, BatchNorm1d
from sorix.nn import Module

def test_linear_initializations():
    # Xavier
    lin_xavier = Linear(10, 5, init='xavier')
    assert lin_xavier.std_dev == np.sqrt(2.0 / (10 + 5))
    
    # He (default)
    lin_he = Linear(10, 5, init='he')
    assert lin_he.std_dev == np.sqrt(2.0 / 10)
    
    # Invalid
    with pytest.raises(ValueError, match="Invalid initialization method"):
        Linear(10, 5, init='invalid')

def test_linear_no_bias():
    lin = Linear(10, 5, bias=False)
    assert lin.b is None
    assert len(lin.parameters()) == 1
    
    x = tensor(np.random.randn(2, 10))
    out = lin(x)
    assert out.shape == (2, 5)

def test_linear_to_cpu():
    lin = Linear(10, 5)
    lin.to('cpu')
    assert lin.device == 'cpu'
    assert lin.W.device == 'cpu'

def test_batchnorm1d_training_eval():
    bn = BatchNorm1d(5)
    assert bn.training == True
    
    x = tensor(np.random.randn(10, 5))
    out_train = bn(x)
    assert out_train.shape == (10, 5)
    
    # Check running stats updated
    assert not np.all(bn.running_mean == 0)
    assert not np.all(bn.running_var == 1)
    
    bn.training = False
    out_eval = bn(x)
    assert out_eval.shape == (10, 5)

def test_batchnorm1d_to_cpu():
    bn = BatchNorm1d(5)
    bn.to('cpu')
    assert bn.device == 'cpu'

def test_neural_network_modes():
    class MultiLayerNet(Module):
        def __init__(self):
            super().__init__()
            self.layers = [Linear(10, 5), BatchNorm1d(5)]
            self.dict_layers = {"l1": Linear(5, 2)}
            
        def forward(self, x):
            x = self.layers[0](x)
            x = self.layers[1](x)
            x = self.dict_layers["l1"](x)
            return x
            
    net = MultiLayerNet()
    
    # Test train mode
    net.train()
    assert net.layers[1].training == True
    
    # Test eval mode
    net.eval()
    assert net.layers[1].training == False
    
    # Test parameters with list and dict
    params = net.parameters()
    # 2 (Linear0) + 2 (BatchNorm) + 2 (Linear1) = 6
    assert len(params) == 6
    
    # Test to() with complex structures
    net.to('cpu')
    assert net.device == 'cpu'

def test_neural_network_state_dict_io():
    class SimpleNet(Module):
        def __init__(self):
            super().__init__()
            self.l1 = Linear(2, 1)
        def forward(self, x): return self.l1(x)
        
    net = SimpleNet()
    w = net.state_dict()
    assert "l1.W" in w
    assert "l1.b" in w
    
    new_net = SimpleNet()
    # Change some values
    new_net.l1.W.data.fill(10.0)
    new_net.load_state_dict(w)
    
    # Check if loaded correctly
    assert np.all(new_net.l1.W.data == net.l1.W.data)

def test_sorix_serialization(tmp_path):
    import sorix
    import os
    
    class SimpleNet(Module):
        def __init__(self):
            super().__init__()
            self.l1 = Linear(2, 1)
        def forward(self, x): return self.l1(x)
        
    net = SimpleNet()
    path = os.path.join(tmp_path, "model.sor")
    sorix.save(net.state_dict(), path)
    
    loaded_state = sorix.load(path)
    assert "l1.W" in loaded_state
    
    new_net = SimpleNet()
    new_net.load_state_dict(loaded_state)
    assert np.all(new_net.l1.W.data == net.l1.W.data)


def test_neural_network_abstract_error():
    net = Module()
    with pytest.raises(NotImplementedError):
        net.forward(tensor([1, 2]))

def test_linear_properties():
    lin = Linear(2, 1)
    lin.W.data = np.array([[1.0], [2.0]])
    lin.b.data = np.array([[0.5]])
    
    assert np.array_equal(lin.coef_, [1.0, 2.0])
    assert lin.intercept_ == 0.5



import numpy as np

class ColumnTransformer:
    def __init__(self, transformers: list[tuple]):
        self.transformers = transformers
        self.n_features = 0
        self.features_names = []


    def fit_transform(self, X):
        
        X_final = np.zeros((len(X), 1))

        for transformer in self.transformers:
            tf = transformer[1]
            features = transformer[2]

            Xt = tf.fit_transform(X[features])
            self.n_features += tf.n_features
            X_final = np.hstack([X_final,Xt])



        return X_final[:,1:]
    
    def transform(self, X):

        X_final = np.zeros((len(X), 1))

        for transformer in self.transformers:
            tf = transformer[1]
            features = transformer[2]

            Xt = tf.transform(X[features])
            X_final = np.hstack([X_final,Xt])



        return X_final[:,1:]
    

    def get_features_names(self):
        
        features = []

        for transformer in self.transformers:
            cat = transformer[0]
            tf = transformer[1]

            features.extend([f'{cat}_{feature}' for feature in tf.get_features_names()])

        return features

    def state_dict(self):
        """Devuelve un diccionario con el estado del transformador de columnas."""
        state = {
            'n_features': self.n_features,
            'features_names': self.features_names,
            'transformers_states': []
        }
        for name, tf, cols in self.transformers:
            if hasattr(tf, 'state_dict') and callable(tf.state_dict):
                state['transformers_states'].append((name, tf.state_dict(), cols))
            else:
                # Fallback if the transformer doesn't have state_dict (unlikely in sorix)
                state['transformers_states'].append((name, tf, cols))
        return state

    def load_state_dict(self, state_dict):
        """Carga el estado del transformador de columnas."""
        self.n_features = state_dict['n_features']
        self.features_names = state_dict['features_names']
        
        # Mapping for easier lookup
        tf_states = {name: (state, cols) for name, state, cols in state_dict['transformers_states']}
        
        for name, tf, cols in self.transformers:
            if name in tf_states:
                state, _ = tf_states[name]
                if hasattr(tf, 'load_state_dict') and callable(tf.load_state_dict):
                    tf.load_state_dict(state)
                else:
                    # If it was saved as the object itself
                    # (this might happen if tf didn't have state_dict when saved)
                    # But in our new system they will have it.
                    pass
        return self




from .metrics import (r2_score, root_mean_squared_error, mean_squared_error, 
                       mean_absolute_error, mean_absolute_percentage_error,
                       accuracy_score, precision_score, recall_score, f1_score,
                       confusion_matrix, classification_report, regression_report)


import numpy as np
from sorix.tensor import Tensor, tensor
from sorix.cupy.cupy import _cupy_available

if _cupy_available:
    import cupy as cp


class MSELoss:
    def __call__(self, y_pred: Tensor, y_real: Tensor):
        return ((y_pred - y_real)**2).mean()
    

class BCEWithLogitsLoss:
    def __call__(self, y_pred: Tensor, y_real: Tensor) -> Tensor:
        xp = cp if y_pred.device == 'gpu' else np

        batch_size = y_real.data.shape[0]

        probs = 1 / (1 + xp.exp(-y_pred.data))
        loss_val = -xp.mean(y_real.data * xp.log(probs) + (1 - y_real.data) * xp.log(1 - probs))
        out = Tensor(loss_val, (y_pred,), 'BCELossWithLogits',device=y_pred.device,requires_grad=y_pred.requires_grad)
        
        def _backward():
            # Usar other.data para claridad
            y_pred.grad += out.grad * (probs- y_real.data)/batch_size
        out._backward = _backward
        return out

    
class CrossEntropyLoss:
    def __init__(self, one_hot=False):
        self.one_hot = one_hot
        self.xp = np

    def __call__(self, y_pred: Tensor, y_real: Tensor) -> Tensor:

        self.xp = cp if y_pred.device == 'gpu' else np

        # Paso 1: Softmax estable
        exp_logits = self.xp.exp(y_pred.data - self.xp.max(y_pred.data, axis=-1, keepdims=True))
        probs = exp_logits / self.xp.sum(exp_logits, axis=-1, keepdims=True)
        batch_size = y_real.data.shape[0]

        # Paso 2: Calcular la pérdida
        if self.one_hot:
            log_probs = self.xp.log(probs + 1e-9)
            loss_val = -self.xp.mean(self.xp.sum(y_real.data * log_probs, axis=-1))
        else:
            Y_indices = y_real.data.flatten().astype(int)
            correct_log_probs = -self.xp.log(probs[self.xp.arange(batch_size), Y_indices] + 1e-9)
            loss_val = self.xp.mean(correct_log_probs)
        
        # Paso 3: Crear el Tensor de pérdida para el backpropagation
        out = Tensor(loss_val, (y_pred,), 'CrossEntropyLoss',device=y_pred.device,requires_grad=y_pred.requires_grad)

        # Paso 4: Unificar el backpropagation
        def _backward():
            if self.one_hot:
                Y_one_hot = y_real.data
            else:
                Y_one_hot = self.xp.zeros_like(probs)
                Y_one_hot[self.xp.arange(batch_size), y_real.data.flatten().astype(int)] = 1
            
            # La derivada combinada es la misma para ambos casos
            grad_combined = (probs - Y_one_hot) / batch_size
            y_pred.grad += out.grad * grad_combined
            
        out._backward = _backward
        return out


import pytest
import numpy as np
import pandas as pd
from sorix.preprocessing import (
    MinMaxScaler, 
    StandardScaler, 
    RobustScaler, 
    OneHotEncoder, 
    ColumnTransformer
)

def test_min_max_scaler():
    data = np.array([[10, 2], [20, 4], [30, 6]])
    scaler = MinMaxScaler()
    
    # Fit and transform
    scaled = scaler.fit_transform(data)
    
    # Expected: 
    # Col 0: (10-10)/20=0, (20-10)/20=0.5, (30-10)/20=1.0
    # Col 1: (2-2)/4=0, (4-2)/4=0.5, (6-2)/4=1.0
    expected = np.array([[0.0, 0.0], [0.5, 0.5], [1.0, 1.0]])
    assert np.allclose(scaled, expected)
    
    # Inverse transform
    original = scaler.inverse_transform(scaled)
    assert np.allclose(original, data)

def test_standard_scaler():
    data = np.array([[1.0], [2.0], [3.0]])
    scaler = StandardScaler()
    
    scaled = scaler.fit_transform(data)
    
    mean = 2.0
    std = np.std([1.0, 2.0, 3.0])
    expected = (data - mean) / std
    
    assert np.allclose(scaled, expected)
    
    original = scaler.inverse_transform(scaled)
    assert np.allclose(original, data)

def test_robust_scaler():
    # Robust to outliers
    data = np.array([[1.0], [2.0], [3.0], [100.0]])
    scaler = RobustScaler()
    
    scaled = scaler.fit_transform(data)
    
    median = np.median(data)
    q1, q3 = np.percentile(data, [25, 75])
    iqr = q3 - q1
    expected = (data - median) / iqr
    
    assert np.allclose(scaled, expected)

def test_one_hot_encoder_numpy():
    data = np.array([['A'], ['B'], ['A'], ['C']])
    # Note: OneHotEncoder in sorix expects a 2D array or DataFrame for categorical_features check
    df = pd.DataFrame(data, columns=['cat'])
    encoder = OneHotEncoder()
    
    encoded = encoder.fit_transform(df)
    
    # Unique values: A, B, C
    # Rows:
    # A -> [1, 0, 0]
    # B -> [0, 1, 0]
    # A -> [1, 0, 0]
    # C -> [0, 0, 1]
    expected = np.array([
        [1.0, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        [1.0, 0.0, 0.0],
        [0.0, 0.0, 1.0]
    ])
    assert np.array_equal(encoded, expected)
    
    encoded_names = encoder.get_features_names()
    assert encoded_names == ['cat_A', 'cat_B', 'cat_C']

def test_column_transformer():
    df = pd.DataFrame({
        'num': [10, 20, 30],
        'cat': ['A', 'B', 'A']
    })
    
    ct = ColumnTransformer([
        ('scale', MinMaxScaler(), ['num']),
        ('encode', OneHotEncoder(), ['cat'])
    ])
    
    result = ct.fit_transform(df)
    
    # num scaled: [0, 0.5, 1.0]
    # cat encoded: A->[1, 0], B->[0, 1], A->[1, 0]
    # Joined: 
    # [0.0, 1.0, 0.0]
    # [0.5, 0.0, 1.0]
    # [1.0, 1.0, 0.0]
    expected = np.array([
        [0.0, 1.0, 0.0],
        [0.5, 0.0, 1.0],
        [1.0, 1.0, 0.0]
    ])
    
    assert np.allclose(result, expected)
    
def test_scalers_edge_cases():
    data = np.array([[1.0], [1.0]]) # Zero variance
    
    # MinMaxScaler zero div
    mms = MinMaxScaler()
    mms.fit(data)
    scaled = mms.transform(data)
    assert np.all(scaled == 0)
    
    # StandardScaler zero div
    ss = StandardScaler()
    ss.fit(data)
    scaled = ss.transform(data)
    assert np.all(scaled == 0)
    
    # RobustScaler zero div
    rs = RobustScaler()
    rs.fit(data)
    scaled = rs.transform(data)
    assert np.all(scaled == 0)

    # Not fitted error
    with pytest.raises(ValueError, match="Debe llamar a fit"):
        MinMaxScaler().transform(data)
    
    # Invalid input type
    with pytest.raises(TypeError, match="La entrada debe ser"):
        MinMaxScaler().fit(123)

    # __repr__
    assert "MinMaxScaler" in repr(mms)

def test_column_transformer_transform():
    df = pd.DataFrame({'num': [10, 20], 'cat': ['A', 'B']})
    df2 = pd.DataFrame({'num': [30], 'cat': ['A']})
    
    ct = ColumnTransformer([
        ('s', MinMaxScaler(), ['num']),
        ('e', OneHotEncoder(), ['cat'])
    ])
    ct.fit_transform(df)
    
    res = ct.transform(df2)
    assert res.shape == (1, 3) # num + 2 cats

def test_one_hot_extra():
    encoder = OneHotEncoder()
    df = pd.DataFrame({'a': ['x', 'y']})
    encoder.fit(df)
    assert "OneHotEncoder" in str(encoder)
    assert "OneHotEncoder" in repr(encoder)
    
    encoded = np.array([[1.0, 0.0], [0.0, 1.0]])
    inv = encoder.inverse_transform(encoded)
    assert np.all(inv == [0, 1])



import numpy as np
import pandas as pd

class OneHotEncoder:
    def __init__(self):
        self.n_features = 0
        self.features = {

        }

    def fit(self, X:pd.DataFrame|np.ndarray):
        
        self.n_features = X.shape[1] if X.ndim == 2 else 1
        categorical_features = X.columns if isinstance(X, pd.DataFrame) else [f"F{i}" for i in range(self.n_features)]

        for feature in categorical_features:
            unique_values = np.unique(X[feature])
            self.features[feature] = unique_values
        return self
    
    def transform(self, X:pd.DataFrame|np.ndarray):


        X_one_hot = np.zeros((len(X),1))

        for feature in self.features.keys():
            X_one_hot = np.hstack([X_one_hot, np.eye(len(self.features[feature]))[np.searchsorted(self.features[feature], X[feature])]])

        self.n_features = X_one_hot.shape[1]-1
        return X_one_hot[:,1:]
    
    def fit_transform(self, X:np.ndarray|pd.DataFrame):
    
        return self.fit(X).transform(X)
        
    def inverse_transform(self, X):
        return np.argmax(X, axis=1)
    

    def get_features_names(self):

        features = []

        for feature,cats in self.features.items():
            for cat in cats:
                features.append(f"{feature}_{cat}")

        return features
            
            
    

    def state_dict(self):
        """Devuelve un diccionario con el estado del codificador."""
        return {k: v for k, v in self.__dict__.items() if not k.startswith('_')}

    def load_state_dict(self, state_dict):
        """Carga el estado del codificador desde un diccionario."""
        for k, v in state_dict.items():
            setattr(self, k, v)
        return self

    def __str__(self) -> str:
        return f"OneHotEncoder({self.features})"
    
    def __repr__(self) -> str:
        return self.__str__()


import pytest
import numpy as np
from sorix import Tensor, tensor
from sorix.clustering.k_means import Kmeans

def test_kmeans_basic_fit():
    # 3 clusters of points
    X_data = np.vstack([
        np.random.randn(10, 2) + np.array([5, 5]),
        np.random.randn(10, 2) + np.array([-5, -5]),
        np.random.randn(10, 2) + np.array([5, -5])
    ])
    X = tensor(X_data)
    
    model = Kmeans(n_clusters=3)
    model.fit(X, max_iters=10)
    
    assert model.centroids.shape == (3, 2)
    assert model.labels.shape == (30,)
    
    # Predict should return a tensor of labels
    preds = model.predict(X)
    assert isinstance(preds, Tensor)
    assert preds.shape == (30,)
    
    # Inertia is a float
    inertia = model.get_inertia(X)
    assert isinstance(inertia, float)
    assert inertia >= 0

def test_kmeans_empty_cluster_handling():
    # Test if it crashes with very few points
    X = tensor([[1.0, 1.0], [2.0, 2.0]])
    model = Kmeans(n_clusters=5) # More clusters than points
    model.fit(X, max_iters=2)
    # Should not crash due to counts[counts == 0] = 1 in _new_centroids
    assert model.centroids.shape == (5, 2)



from sorix.tensor import Tensor, tensor
import numpy as np
from sorix.cupy.cupy import _cupy_available

if _cupy_available:
    import cupy as cp


class Optimizer:

    def __init__(self, parameters: list[Tensor], lr=1e-3):
        self.parameters = parameters
        self.lr = lr
        self.device = parameters[0].device
        self.xp = cp if self.device == 'gpu' else np

    
    def zero_grad(self):

        for param in self.parameters:
            param.grad = self.xp.zeros_like(param.grad)
    
    def step(self):
        raise NotImplementedError
    


class SGD(Optimizer):
    def __init__(self, parameters: list[Tensor], lr=1e-3):
        super().__init__(parameters, lr)


    def step(self):
        for param in self.parameters:
            param.data -= self.lr * param.grad


class SGDMomentum(Optimizer):
    def __init__(self, parameters: list[Tensor], lr=1e-3, momentum=0.9):
        super().__init__(parameters, lr)
        self.momentum = momentum
        self.vts = {}

    def step(self):

        for i, param in enumerate(self.parameters):
            if i not in self.vts:
                self.vts[i] = self.xp.zeros_like(param.data)
            self.vts[i] = self.momentum * self.vts[i] + param.grad
            param.data -= self.lr * self.vts[i]


class RMSprop(Optimizer):
    def __init__(self, parameters: list[Tensor], lr=1e-3, decay=0.9,epsilon=1e-8):
        super().__init__(parameters, lr)
        self.decay = decay
        self.vts = {}
        self.epsilon = epsilon

    def step(self):

        for i, param in enumerate(self.parameters):
            if i not in self.vts:
                self.vts[i] = self.xp.zeros_like(param.data)
            self.vts[i] = self.decay * self.vts[i] + (1 - self.decay) * param.grad ** 2
            param.data -= self.lr * param.grad / (self.xp.sqrt(self.vts[i]) + self.epsilon)



class Adam(Optimizer):
    def __init__(self, parameters: list[Tensor], lr=1e-3, beta1=0.9, beta2=0.999, epsilon=1e-8):
        super().__init__(parameters, lr)
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.vts = {}
        self.rts = {}
        self.t = 0

    def step(self):
        self.t += 1

        for i, param in enumerate(self.parameters):
            if i not in self.vts:
                self.vts[i] = self.xp.zeros_like(param.data)
                self.rts[i] = self.xp.zeros_like(param.data)
            self.vts[i] = self.beta1 * self.vts[i] + (1 - self.beta1) * param.grad
            self.rts[i] = self.beta2 * self.rts[i] + (1 - self.beta2) * param.grad ** 2
            v_hat = self.vts[i] / (1 - self.beta1 ** self.t)
            r_hat = self.rts[i] / (1 - self.beta2 ** self.t)

            param.data -= self.lr * v_hat / (self.xp.sqrt(r_hat) + self.epsilon)


import pytest
import numpy as np
from sorix import Tensor, tensor
from sorix.nn import MSELoss, BCEWithLogitsLoss, CrossEntropyLoss

def test_mse_loss():
    """Test Mean Squared Error loss and its gradient."""
    y_pred = tensor([[2.0], [4.0]], requires_grad=True)
    y_true = tensor([[1.0], [5.0]])
    
    criterion = MSELoss()
    loss = criterion(y_pred, y_true)
    
    # Calculation: ((2-1)^2 + (4-5)^2) / 2 = (1 + 1) / 2 = 1.0
    assert np.isclose(loss.data, 1.0)
    
    loss.backward()
    # dLoss/dy_pred = 2 * (y_pred - y_true) / size
    # For element 0: 2 * (2-1) / 2 = 1.0
    # For element 1: 2 * (4-5) / 2 = -1.0
    expected_grad = np.array([[1.0], [-1.0]])
    assert np.allclose(y_pred.grad, expected_grad)

def test_bce_with_logits_loss():
    """Test Binary Cross Entropy with Logits and its gradient."""
    # Logits (before sigmoid)
    y_pred = tensor([0.0, 2.0, -2.0], requires_grad=True)
    y_true = tensor([0.0, 1.0, 0.0])
    
    criterion = BCEWithLogitsLoss()
    loss = criterion(y_pred, y_true)
    
    # Sigmoids: [0.5, 0.8808, 0.1192]
    # Loss: -1/3 * [ (0*log(0.5) + 1*log(0.5)) + (1*log(0.88) + 0*log(0.12)) + (0*log(0.11) + 1*log(0.88)) ]
    # Manual check: sig = 1/(1+exp(-x)); loss = -mean(y*log(sig) + (1-y)*log(1-sig))
    probs = 1 / (1 + np.exp(-y_pred.data))
    expected_loss = -np.mean(y_true.data * np.log(probs) + (1 - y_true.data) * np.log(1 - probs))
    
    assert np.allclose(loss.data, expected_loss)
    
    loss.backward()
    # Gradient for BCEWithLogits: (sigmoid(y_pred) - y_true) / batch_size
    expected_grad = (probs - y_true.data) / 3.0
    assert np.allclose(y_pred.grad, expected_grad)

def test_cross_entropy_loss_indices():
    """Test CrossEntropyLoss with integer indices."""
    # Batch of 2, 3 classes
    y_pred = tensor([[1.0, 2.0, 3.0], 
                     [5.0, 2.0, 1.0]], requires_grad=True)
    y_true = tensor([2, 0]) # Targets: Class 2 for sample 0, Class 0 for sample 1
    
    criterion = CrossEntropyLoss(one_hot=False)
    loss = criterion(y_pred, y_true)
    
    # Softmax
    exp_logits = np.exp(y_pred.data - np.max(y_pred.data, axis=-1, keepdims=True))
    probs = exp_logits / np.sum(exp_logits, axis=-1, keepdims=True)
    
    # Loss: -mean(log(probs_at_target))
    expected_loss = -np.mean(np.log(np.array([probs[0, 2], probs[1, 0]]) + 1e-9))
    
    assert np.allclose(loss.data, expected_loss)
    
    loss.backward()
    # Gradient: (probs - target_one_hot) / batch_size
    target_one_hot = np.zeros_like(probs)
    target_one_hot[0, 2] = 1
    target_one_hot[1, 0] = 1
    expected_grad = (probs - target_one_hot) / 2.0
    
    assert np.allclose(y_pred.grad, expected_grad)

def test_cross_entropy_loss_one_hot():
    """Test CrossEntropyLoss with one-hot encoded targets."""
    y_pred = tensor([[1.0, 0.0], [0.0, 1.0]], requires_grad=True)
    y_true = tensor([[1.0, 0.0], [0.0, 1.0]]) # One-hot
    
    criterion = CrossEntropyLoss(one_hot=True)
    loss = criterion(y_pred, y_true)
    
    # Logic same as above, just checking one_hot flag
    loss.backward()
    assert y_pred.grad.shape == y_pred.shape



import pytest
import numpy as np
import pandas as pd
import os
import sorix
from sorix.preprocessing import MinMaxScaler, StandardScaler, RobustScaler, OneHotEncoder, ColumnTransformer

def test_scaler_save_load(tmp_path):
    # Test MinMaxScaler
    data = np.array([[1], [2], [3], [4], [5]]).astype(float)
    scaler = MinMaxScaler().fit(data)
    path = os.path.join(tmp_path, "minmax.pkl")
    
    # Save whole object (falls back to pickle)
    sorix.save(scaler, path)
    loaded_scaler = sorix.load(path)
    
    assert isinstance(loaded_scaler, MinMaxScaler)
    assert np.allclose(loaded_scaler.transform(data), scaler.transform(data))
    
    # Test state_dict saving (saves as .npz)
    state_path = os.path.join(tmp_path, "scaler_state.sor")
    sorix.save(scaler.state_dict(), state_path)
    
    new_scaler = MinMaxScaler()
    # load returns a dict from .npz
    loaded_state = sorix.load(state_path)
    new_scaler.load_state_dict(loaded_state)
    
    assert np.allclose(new_scaler.transform(data), scaler.transform(data))

def test_encoder_save_load(tmp_path):
    df = pd.DataFrame({'color': ['red', 'blue', 'green', 'red']})
    encoder = OneHotEncoder().fit(df)
    path = os.path.join(tmp_path, "encoder.pkl")
    
    sorix.save(encoder, path)
    loaded_encoder = sorix.load(path)
    
    assert isinstance(loaded_encoder, OneHotEncoder)
    assert np.array_equal(loaded_encoder.transform(df), encoder.transform(df))
    
    # state_dict
    state_path = os.path.join(tmp_path, "encoder_state.sor")
    sorix.save(encoder.state_dict(), state_path)
    
    new_encoder = OneHotEncoder()
    new_encoder.load_state_dict(sorix.load(state_path))
    assert np.array_equal(new_encoder.transform(df), encoder.transform(df))

def test_column_transformer_save_load(tmp_path):
    df = pd.DataFrame({
        'age': [25, 30, 35, 40],
        'city': ['NY', 'LA', 'NY', 'CHI']
    })
    
    ct = ColumnTransformer([
        ('scaler', StandardScaler(), ['age']),
        ('encoder', OneHotEncoder(), ['city'])
    ])
    
    ct.fit_transform(df)
    path = os.path.join(tmp_path, "ct.pkl")
    
    sorix.save(ct, path)
    loaded_ct = sorix.load(path)
    
    assert isinstance(loaded_ct, ColumnTransformer)
    assert np.allclose(loaded_ct.transform(df), ct.transform(df))
    
    # recursive state_dict
    state_path = os.path.join(tmp_path, "ct_state.sor")
    sorix.save(ct.state_dict(), state_path)
    
    new_ct = ColumnTransformer([
        ('scaler', StandardScaler(), ['age']),
        ('encoder', OneHotEncoder(), ['city'])
    ])
    new_ct.load_state_dict(sorix.load(state_path))
    assert np.allclose(new_ct.transform(df), ct.transform(df))



import numpy as np
import pandas as pd
from typing import Union, Optional, List

class BaseScaler:
    """Clase base para todos los escaladores, implementando métodos comunes."""
    def __init__(self):
        self.numerical_features: List[str] = []
        self.n_features: int = 0

    def prepros(self, X: Union[np.ndarray, pd.DataFrame]):
        """Valida y registra nombres de columnas."""
        if isinstance(X, pd.DataFrame):
            self.numerical_features = list(X.columns)
            X = X.to_numpy()
        elif isinstance(X, np.ndarray):
            self.numerical_features = [f"F{i}" for i in range(X.shape[1])] if X.ndim > 1 else ["F0"]
        else:
            raise TypeError("La entrada debe ser un ndarray de NumPy o un DataFrame de Pandas.")
        
        self.n_features = X.shape[1] if X.ndim > 1 else 1
        return X

    def fit(self, X: Union[np.ndarray, pd.DataFrame]):
        raise NotImplementedError

    def transform(self, X: Union[np.ndarray, pd.DataFrame]):
        raise NotImplementedError

    def fit_transform(self, X: Union[np.ndarray, pd.DataFrame]):
        self.fit(X)
        return self.transform(X)

    def inverse_transform(self, X: Union[np.ndarray, pd.DataFrame]):
        raise NotImplementedError
    
    def get_features_names(self):
        return self.numerical_features

    def state_dict(self):
        """Devuelve un diccionario con el estado del escalador."""
        return {k: v for k, v in self.__dict__.items() if not k.startswith('_')}

    def load_state_dict(self, state_dict):
        """Carga el estado del escalador desde un diccionario."""
        for k, v in state_dict.items():
            setattr(self, k, v)
        return self

    def __repr__(self):
        params = ", ".join(f"{k}={v}" for k, v in self.__dict__.items())
        return f"{self.__class__.__name__}({params})"



class MinMaxScaler(BaseScaler):
    """Escala las características a un rango [0, 1]."""
    def __init__(self):
        super().__init__()
        self.min: Optional[np.ndarray] = None
        self.max: Optional[np.ndarray] = None

    def fit(self, X):
        X = self.prepros(X)
        self.min = X.min(axis=0)
        self.max = X.max(axis=0)
        return self

    def transform(self, X):
        X = self.prepros(X)  # asegura nombres si no se hizo antes
        if self.min is None or self.max is None:
            raise ValueError("Debe llamar a fit() antes de transform().")
        denom = self.max - self.min
        denom[denom == 0] = 1e-9
        return (X - self.min) / denom

    def inverse_transform(self, X):
        X = self.prepros(X)
        return X * (self.max - self.min) + self.min


class StandardScaler(BaseScaler):
    """Estandariza eliminando la media y escalando a varianza unitaria."""
    def __init__(self):
        super().__init__()
        self.mean: Optional[np.ndarray] = None
        self.std: Optional[np.ndarray] = None

    def fit(self, X):
        X = self.prepros(X)
        self.mean = X.mean(axis=0)
        self.std = X.std(axis=0)
        return self

    def transform(self, X):
        X = self.prepros(X)
        if self.mean is None or self.std is None:
            raise ValueError("Debe llamar a fit() antes de transform().")
        std_safe = self.std.copy()
        std_safe[std_safe == 0] = 1e-9
        return (X - self.mean) / std_safe

    def inverse_transform(self, X):
        X = self.prepros(X)
        return X * self.std + self.mean


class RobustScaler(BaseScaler):
    """Escala usando mediana e IQR (robusto a outliers)."""
    def __init__(self):
        super().__init__()
        self.median: Optional[np.ndarray] = None
        self.q1: Optional[np.ndarray] = None
        self.q3: Optional[np.ndarray] = None

    def fit(self, X):
        X = self.prepros(X)
        self.median = np.quantile(X, 0.5, axis=0)
        self.q1 = np.quantile(X, 0.25, axis=0)
        self.q3 = np.quantile(X, 0.75, axis=0)
        return self

    def transform(self, X):
        X = self.prepros(X)
        if self.median is None:
            raise ValueError("Debe llamar a fit() antes de transform().")
        iqr = self.q3 - self.q1
        iqr[iqr == 0] = 1e-9
        return (X - self.median) / iqr

    def inverse_transform(self, X):
        X = self.prepros(X)
        return X * (self.q3 - self.q1) + self.median



class Dataset:
    def __init__(self, X, y=None):
        if y is not None and len(X) != len(y):
            raise ValueError(f"X and y must have the same length. Got len(X)={len(X)} and len(y)={len(y)}")
        self.X = X
        self.y = y

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        if self.y is not None:
            return self.X[idx], self.y[idx]
        return self.X[idx]

    def __setitem__(self, idx, value):
        if self.y is not None:
            self.X[idx], self.y[idx] = value
        else:
            self.X[idx] = value

    def __str__(self):
        return f"Dataset(len={len(self)}, has_labels={self.y is not None})"

    def __repr__(self):
        return self.__str__()


import pytest
import numpy as np
from sorix import Tensor, tensor
from sorix.optim import SGD, SGDMomentum, RMSprop, Adam

def test_optimizer_zero_grad():
    params = [tensor([1.0, 2.0], requires_grad=True)]
    params[0].grad = np.array([0.5, 0.5])
    
    optimizer = SGD(params, lr=0.1)
    optimizer.zero_grad()
    
    assert np.all(params[0].grad == 0.0)

def test_sgd_step():
    param = tensor([10.0], requires_grad=True)
    param.grad = np.array([2.0]) # Positive gradient
    
    optimizer = SGD([param], lr=0.1)
    optimizer.step()
    
    # new_data = 10.0 - 0.1 * 2.0 = 9.8
    assert np.allclose(param.data, [9.8])

def test_sgd_momentum_step():
    param = tensor([10.0], requires_grad=True)
    param.grad = np.array([2.0])
    
    # v = 0.9 * 0 + 2.0 = 2.0
    # data = 10.0 - 0.1 * 2.0 = 9.8
    optimizer = SGDMomentum([param], lr=0.1, momentum=0.9)
    optimizer.step()
    assert np.allclose(param.data, [9.8])
    
    # Step 2:
    # v = 0.9 * 2.0 + 2.0 = 3.8
    # data = 9.8 - 0.1 * 3.8 = 9.42
    optimizer.step()
    assert np.allclose(param.data, [9.42])

def test_rmsprop_step():
    param = tensor([10.0], requires_grad=True)
    param.grad = np.array([2.0])
    
    # lr=0.1, decay=0.9, eps=1e-8
    # vt = 0.9 * 0 + 0.1 * (2.0^2) = 0.4
    # data = 10.0 - 0.1 * 2.0 / sqrt(0.4) = 10.0 - 0.2 / 0.63245 = 10.0 - 0.31622 = 9.68377
    optimizer = RMSprop([param], lr=0.1, decay=0.9)
    optimizer.step()
    assert np.allclose(param.data, [9.68377], atol=1e-5)

def test_adam_step():
    param = tensor([10.0], requires_grad=True)
    param.grad = np.array([2.0])
    
    # beta1=0.9, beta2=0.999
    # t=1
    # v = 0.1 * 2.0 = 0.2
    # r = 0.001 * 4.0 = 0.004
    # v_hat = 0.2 / (1 - 0.9) = 2.0
    # r_hat = 0.004 / (1 - 0.999) = 4.0
    # data = 10.0 - 0.1 * 2.0 / (sqrt(4.0) + 1e-8) = 10.0 - 0.1 * 2.0 / 2.0 = 9.9
    optimizer = Adam([param], lr=0.1)
    optimizer.step()
    assert np.allclose(param.data, [9.9])

def test_optimization_loop():
    """Test if SGD can minimize a simple quadratic function."""
    x = tensor([10.0], requires_grad=True)
    optimizer = SGD([x], lr=0.1)
    
    # Minimize f(x) = x^2
    for _ in range(50):
        optimizer.zero_grad()
        loss = x * x
        loss.backward()
        optimizer.step()
    
    # After 50 steps with lr=0.1, x should be very close to 0
    # x_new = x - 0.1 * 2x = 0.8x
    # x_50 = 10 * (0.8^50) approx 0
    assert abs(x.data[0]) < 1e-3



import pandas as pd
import numpy as np

def train_test_split(X: pd.DataFrame,
                     Y: pd.Series = None,
                     test_size: float = 0.2,
                     random_state: int = 42,
                     shuffle: bool = True) -> tuple:
    """
    Method to split the data into train and test sets.

    Args:
        X (pd.DataFrame): Features.
        Y (pd.Series): Labels.
        test_size (float, optional): Proportion of the dataset to include in the test split. Default is 0.2.
        random_state (int, optional): Random state for reproducibility. Default is 42.
        shuffle (bool, optional): Whether to shuffle the data before splitting. Default is True.

    Returns:
        tuple: (X_train, X_test, y_train, y_test)
    """
    
    # Validate test_size
    if not (0 < test_size < 1):
        raise ValueError("test_size must be between 0 and 1.")

    # Set random seed for reproducibility
    np.random.seed(random_state)

    # Shuffle the data if specified
    if shuffle:
        indices = np.random.permutation(len(X))
        X = X.iloc[indices]
        if Y is not None:
            Y = Y.iloc[indices]

    # Calculate the split index
    split_index = int(len(X) * (1 - test_size))

    # Split the data
    X_train = X.iloc[:split_index]
    X_test = X.iloc[split_index:]
    if Y is not None:
        Y_train = Y.iloc[:split_index]
        Y_test = Y.iloc[split_index:]
        return X_train, X_test, Y_train, Y_test

    return X_train, X_test



from sorix.cupy.cupy import _cupy_available

def is_available(verbose: bool = True) -> bool:


    if not _cupy_available:
        if verbose:
            print("❌ CuPy is not installed.")
        return False

    import cupy as cp
    
    try:
        n_gpus = cp.cuda.runtime.getDeviceCount()
        if n_gpus == 0:
            if verbose:
                print("❌ No CUDA devices found.")
            return False

        # Test 1: operación simple en GPU (memoria + aritmética básica)
        try:
            _ = int((cp.arange(5) * 2).sum())
            print("✅ GPU basic operation passed")
        except Exception as e:
            if verbose:
                print("❌ GPU basic operation failed:", e)
            return False

        # Test 2: operación que fuerza cuBLAS (matmul)
        try:
            A = cp.random.rand(4, 4)
            B = cp.random.rand(4, 4)
            _ = int((A @ B).sum())  # fuerza cuBLAS
        except Exception as e:
            if verbose:
                print("❌ cuBLAS/linear algebra operation failed:", e)
            return False

        if verbose:
            props = cp.cuda.runtime.getDeviceProperties(0)
            print(f"✅ GPU available: {props['name'].decode('utf-8')}")
            print(f"CUDA runtime version: {cp.cuda.runtime.runtimeGetVersion()}")
            print(f"CuPy version: {cp.__version__}")

        return True

    except Exception as e:
        if verbose:
            print("❌ Error while checking CuPy/CUDA:", e)
        return False





import pytest
import numpy as np
from sorix import Tensor, tensor, no_grad

def test_tensor_initialization():
    """Test standard tensor creation and properties."""
    data = [1, 2, 3]
    t = tensor(data)
    assert np.array_equal(t.data, np.array(data))
    assert t.device == 'cpu'
    assert t.requires_grad == False
    assert t.grad is None

    t_grad = tensor(data, requires_grad=True)
    assert t_grad.requires_grad == True
    assert t_grad.grad is not None
    assert np.all(t_grad.grad == 0)

def test_basic_arithmetic_autograd():
    """Test +, -, *, / and their gradients."""
    a = tensor([2.0, 3.0], requires_grad=True)
    b = tensor([4.0, 5.0], requires_grad=True)
    
    # Addition
    c = a + b
    c.backward()
    assert np.array_equal(c.data, [6.0, 8.0])
    assert np.array_equal(a.grad, [1.0, 1.0])
    assert np.array_equal(b.grad, [1.0, 1.0])
    
    a.grad = None; b.grad = None
    
    # Subtraction
    d = a - b
    d.backward()
    assert np.array_equal(d.data, [-2.0, -2.0])
    assert np.array_equal(a.grad, [1.0, 1.0])
    assert np.array_equal(b.grad, [-1.0, -1.0])
    
    a.grad = None; b.grad = None

    # Multiplication
    e = a * b
    e.backward()
    assert np.array_equal(e.data, [8.0, 15.0])
    assert np.array_equal(a.grad, [4.0, 5.0]) # de/da = b
    assert np.array_equal(b.grad, [2.0, 3.0]) # de/db = a

    a.grad = None; b.grad = None
    
    # Division
    f = a / b
    f.backward()
    assert np.allclose(f.data, [0.5, 0.6])
    assert np.allclose(a.grad, [1/4.0, 1/5.0]) # df/da = 1/b
    assert np.allclose(b.grad, [-2.0/16.0, -3.0/25.0]) # df/db = -a/b^2

def test_matmul_autograd():
    """Test matrix multiplication and its gradients."""
    # (2, 3) @ (3, 2) -> (2, 2)
    a_data = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    b_data = np.array([[7.0, 8.0], [9.0, 10.0], [11.0, 12.0]])
    
    a = tensor(a_data, requires_grad=True)
    b = tensor(b_data, requires_grad=True)
    
    c = a @ b
    c.backward()
    
    expected_c = a_data @ b_data
    assert np.array_equal(c.data, expected_c)
    
    # dc/da = grad_out @ b.T. Since grad_out is ones(2,2)
    # grad_a = ones(2,2) @ b.T
    expected_grad_a = np.ones((2, 2)) @ b_data.T
    expected_grad_b = a_data.T @ np.ones((2, 2))
    
    assert np.allclose(a.grad, expected_grad_a)
    assert np.allclose(b.grad, expected_grad_b)

def test_nonlinear_functions():
    """Test tanh and sigmoid and their gradients."""
    x = tensor([0.0, 1.0, -1.0], requires_grad=True)
    
    # Tanh
    y = x.tanh()
    y.backward()
    expected_y = np.tanh(x.data)
    assert np.allclose(y.data, expected_y)
    # dy/dx = 1 - tanh^2
    expected_grad = 1 - expected_y**2
    assert np.allclose(x.grad, expected_grad)

    x.grad = None
    # Sigmoid
    z = x.sigmoid()
    z.backward()
    expected_z = 1 / (1 + np.exp(-x.data))
    assert np.allclose(z.data, expected_z)
    # dz/dx = s * (1 - s)
    expected_grad_z = expected_z * (1 - expected_z)
    assert np.allclose(x.grad, expected_grad_z)

def test_reduction_ops():
    """Test sum and mean."""
    x_data = np.array([[1.0, 2.0], [3.0, 4.0]])
    x = tensor(x_data, requires_grad=True)
    
    # Sum
    s = x.sum()
    s.backward()
    assert s.data == 10.0
    assert np.array_equal(x.grad, np.ones((2, 2)))
    
    x.grad = None
    # Mean
    m = x.mean()
    m.backward()
    assert m.data == 2.5
    assert np.allclose(x.grad, np.full((2, 2), 1.0/4.0))

def test_broadcasting_autograd():
    """Test if gradients are correctly matched during broadcasting."""
    a = tensor([[1.0, 2.0], [3.0, 4.0]], requires_grad=True) # (2, 2)
    b = tensor([10.0, 20.0], requires_grad=True)            # (2,) -> broadcast to (2, 2)
    
    c = a + b
    c.backward()
    
    assert np.array_equal(a.grad, np.ones((2, 2)))
    # b is broadcasted, so its gradient should be the sum over the broadcasted dimension
    assert np.array_equal(b.grad, [2.0, 2.0])

def test_no_grad():
    """Test the no_grad context manager."""
    a = tensor([1.0, 2.0], requires_grad=True)
    b = tensor([3.0, 4.0], requires_grad=True)
    
    with no_grad():
        c = a + b
    
    assert c.requires_grad == False
    assert len(c._prev) == 0

def test_complex_expression():
    """A more complex expression to verify chain rule."""
    x = tensor(2.0, requires_grad=True)
    y = tensor(3.0, requires_grad=True)
    
    # f(x, y) = x^2 * y + y^2
    f = (x**2) * y + (y**2)
    f.backward()
    
    # df/dx = 2xy = 2*2*3 = 12
    # df/dy = x^2 + 2y = 2^2 + 2*3 = 4 + 6 = 10
    assert x.grad == 12.0
    assert y.grad == 10.0
    # Wait, let's check how sorix handles initialization of grad in backward.



import pytest
import numpy as np
import pandas as pd
from sorix.utils.data import Dataset, DataLoader
from sorix.model_selection import train_test_split


def test_dataset_basics():
    """Test basic Dataset functionality: len and indexing."""
    X = np.array([[1, 2], [3, 4], [5, 6]])
    y = np.array([0, 1, 0])
    
    ds = Dataset(X, y)
    
    assert len(ds) == 3
    
    # Check single item access
    x0, y0 = ds[0]
    assert np.array_equal(x0, [1, 2])
    assert y0 == 0
    
    # Check slice access (returns tuples of arrays)
    tx, ty = ds[0:2]
    assert np.array_equal(tx, [[1, 2], [3, 4]])
    assert np.array_equal(ty, [0, 1])

def test_dataset_extra():
    X = np.array([[1, 2]])
    y = np.array([0])
    ds = Dataset(X, y)
    ds[0] = (np.array([3, 4]), np.array([1]))
    assert np.all(ds.X[0] == [3, 4])
    assert "Dataset" in str(ds)
    assert "Dataset" in repr(ds)

def test_dataloader_batching():
    """Test DataLoader batching logic and length."""
    X = np.arange(10).reshape(10, 1)
    y = np.arange(10)
    ds = Dataset(X, y)
    
    # Case 1: Divisible batch size
    dl = DataLoader(ds, batch_size=5)
    assert len(dl) == 2
    batches = list(dl)
    assert len(batches) == 2
    assert batches[0][0].shape == (5, 1)
    assert batches[1][0].shape == (5, 1)
    
    # Case 2: Non-divisible batch size
    dl2 = DataLoader(ds, batch_size=4)
    assert len(dl2) == 3 # 4+4+2
    batches2 = list(dl2)
    assert len(batches2) == 3
    assert batches2[0][0].shape == (4, 1)
    assert batches2[2][0].shape == (2, 1)

def test_train_test_split_pandas():
    """Test train_test_split with Pandas objects."""
    df = pd.DataFrame({
        'feature': np.arange(100)
    })
    labels = pd.Series(np.arange(100))
    
    X_train, X_test, y_train, y_test = train_test_split(
        df, labels, test_size=0.2, random_state=42, shuffle=True
    )
    
    assert len(X_train) == 80
    assert len(X_test) == 20
    assert len(y_train) == 80
    assert len(y_test) == 20
    
    # Check for shuffling (random_state=42 index 0 usually isn't 0)
    assert X_train.index[0] != 0
    
    # Check alignment (index of X and y must match)
    assert X_train.index[0] == y_train.index[0]

def test_train_test_split_no_labels():
    """Test train_test_split when only X is provided."""
    df = pd.DataFrame({'f': np.arange(10)})
    
    X_train, X_test = train_test_split(df, test_size=0.3, shuffle=False)
    
    assert len(X_train) == 7
    assert len(X_test) == 3
    # No shuffle, so it should be sequential
    assert X_train.index[0] == 0
    assert X_test.index[0] == 7



"""
Macros module for MkDocs to provide dynamic variables.
"""
import os
import subprocess


def define_env(env):
    """
    Define custom variables and macros for MkDocs.
    
    This function is called by mkdocs-macros-plugin.
    """
    
    # Detect the current git branch
    def get_git_branch():
        """Get the current git branch name."""
        # First, try to get from environment variable (set by CI/CD)
        branch = os.environ.get('GIT_BRANCH')
        
        if branch:
            return branch
        
        # If not in CI, try to detect from git
        try:
            result = subprocess.run(
                ['git', 'rev-parse', '--abbrev-ref', 'HEAD'],
                capture_output=True,
                text=True,
                check=True
            )
            return result.stdout.strip()
        except (subprocess.CalledProcessError, FileNotFoundError):
            # Fallback to 'develop' if git is not available
            return 'develop'
    
    # Set the branch variable
    env.variables['branch'] = get_git_branch()
    
    # Also provide the full GitHub URL base for convenience
    env.variables['github_base'] = 'https://github.com/Mitchell-Mirano/sorix'
    env.variables['github_blob'] = f"{env.variables['github_base']}/blob/{env.variables['branch']}"
    
    # Colab base URL
    env.variables['colab_base'] = f"https://colab.research.google.com/github/Mitchell-Mirano/sorix/blob/{env.variables['branch']}"
    
    # Documentation base URL - use GitHub Pages in CI, localhost in local dev
    # Check if we're in CI/CD (GitHub Actions sets CI=true)
    is_ci = os.environ.get('CI', '').lower() == 'true'
    if is_ci:
        env.variables['docs_base'] = 'https://mitchell-mirano.github.io/sorix'
    else:
        env.variables['docs_base'] = 'http://127.0.0.1:8000'



import pytest
import numpy as np
from sorix import Tensor, tensor, no_grad

def test_tensor_inplace():
    a = tensor([1.0, 2.0])
    
    a.add_(1.0)
    assert np.array_equal(a.data, [2.0, 3.0])
    
    a.sub_(1.0)
    assert np.array_equal(a.data, [1.0, 2.0])
    
    a.mul_(2.0)
    assert np.array_equal(a.data, [2.0, 4.0])
    
    a.div_(2.0)
    assert np.array_equal(a.data, [1.0, 2.0])

def test_tensor_reshape_transpose():
    a = tensor([[1, 2, 3], [4, 5, 6]], requires_grad=True)
    
    # Reshape
    b = a.reshape(3, 2)
    assert b.shape == (3, 2)
    b.backward()
    assert a.grad.shape == (2, 3)
    
    a.grad = None
    # Flatten
    c = a.flatten()
    assert c.shape == (6,)
    c.backward()
    assert a.grad.shape == (2, 3)
    
    a.grad = None
    # Transpose
    d = a.transpose(1, 0)
    assert d.shape == (3, 2)
    d.backward()
    assert a.grad.shape == (2, 3)
    
    # T property
    assert a.T.shape == (3, 2)

def test_tensor_abs():
    a = tensor([-1.0, 2.0, -3.0])
    b = a.abs()
    assert np.array_equal(b.data, [1.0, 2.0, 3.0])

def test_tensor_comparisons():
    a = tensor([1, 2, 3])
    b = tensor([2, 2, 2])
    
    assert np.array_equal((a > b).data, [False, False, True])
    assert np.array_equal((a < b).data, [True, False, False])
    assert np.array_equal((a >= b).data, [False, True, True])
    assert np.array_equal((a <= b).data, [True, True, False])
    assert np.array_equal((a == b).data, [False, True, False])
    assert np.array_equal((a != b).data, [True, False, True])
    
    # Comparison with scalar
    assert np.array_equal((a > 2).data, [False, False, True])

def test_tensor_utility_methods():
    a = tensor([1.5, 2.5], requires_grad=True)
    
    # astype
    b = a.astype(int)
    assert b.dtype == int
    
    # item
    c = tensor([5.0])
    assert c.item() == 5.0
    
    # to_numpy
    assert isinstance(a.to_numpy(), np.ndarray)
    
    # __array__
    assert isinstance(np.array(a), np.ndarray)
    
    # __iter__
    items = [x for x in a]
    assert len(items) == 2
    
    # __str__, __repr__
    assert "Tensor" in str(a)
    assert "Tensor" in repr(a)

def test_tensor_getitem_autograd():
    a = tensor([1.0, 2.0, 3.0], requires_grad=True)
    b = a[1]
    assert b.data == 2.0
    b.backward()
    assert np.array_equal(a.grad, [0.0, 1.0, 0.0])

def test_tensor_radd_rsub_rmul_rtruediv_rmatmul():
    a = tensor([2.0], requires_grad=True)
    
    # radd
    b = 1.0 + a
    assert b.data == 3.0
    
    # rsub
    c = 5.0 - a
    assert c.data == 3.0
    c.backward()
    assert a.grad == [-1.0]
    
    a.grad = None
    # rmul
    d = 3.0 * a
    assert d.data == 6.0
    d.backward()
    assert a.grad == [3.0]
    
    a.grad = None
    # rtruediv
    e = 10.0 / a
    assert e.data == 5.0
    e.backward()
    # d/da (10/a) = -10/a^2 = -10/4 = -2.5
    assert a.grad == [-2.5]
    
    # rmatmul
    x = tensor([[1.0, 2.0]])
    y = np.array([[3.0], [4.0]])
    z = y @ x # This uses __rmatmul__ of x
    assert z.shape == (2, 2)


def test_tensor_backward_no_grad_root():
    a = tensor([1.0, 2.0], requires_grad=False)
    a.backward() # Should initialize a.grad to ones
    assert np.array_equal(a.grad, [1.0, 1.0])

def test_tensor_to_gpu_error():
    from sorix.cupy.cupy import _cupy_available
    if not _cupy_available:
        a = tensor([1, 2])
        with pytest.raises(RuntimeError, match="CuPy no está instalado"):
            a.to("gpu")
        with pytest.raises(ValueError, match="device debe ser 'cpu' o 'gpu'"):
            a.to("invalid")

def test_tensor_broadcasting_complex():
    a = tensor(np.ones((2, 1, 4)), requires_grad=True)
    b = tensor(np.ones((1, 3, 4)), requires_grad=True)
    c = a + b # (2, 3, 4)
    c.backward()
    assert a.grad.shape == (2, 1, 4)
    assert b.grad.shape == (1, 3, 4)
    assert np.all(a.grad == 3.0)
    assert np.all(b.grad == 2.0)

def test_tensor_gpu_ops_if_available():
    from sorix.cupy.cupy import _cupy_available
    if _cupy_available:
        import cupy as cp
        a = tensor([1.0, 2.0], device='gpu', requires_grad=True)
        b = tensor([3.0, 4.0], device='gpu', requires_grad=True)
        
        # Arithmetic
        c = a + b
        assert c.device == 'gpu'
        assert isinstance(c.data, cp.ndarray)
        
        # Non-linear with backward
        d = a.tanh()
        assert d.device == 'gpu'
        d.sum().backward()
        
        a.grad = cp.zeros_like(a.data)
        e = a.sigmoid()
        assert e.device == 'gpu'
        e.sum().backward()
        
        a.grad = cp.zeros_like(a.data)
        f = a.softmax()
        assert f.device == 'gpu'
        f.sum().backward()
        
        # Power
        p = a**2
        p.sum().backward()



import numpy as np
import pandas as pd
from sorix.cupy.cupy import _cupy_available

if _cupy_available:
    import cupy as cp


_autograd_enabled = True

class no_grad:
    def __init__(self):
        self.prev = True

    def __enter__(self):
        global _autograd_enabled
        self.prev = _autograd_enabled
        _autograd_enabled = False

    def __exit__(self, exc_type, exc_val, exc_tb):
        global _autograd_enabled
        _autograd_enabled = self.prev

def _noop():
    """Función vacía para usar como backward por defecto."""
    return None


class Tensor:

    def __init__(self, data, _children=[], _op='',device='cpu',requires_grad=False):

        
        if device == 'gpu' and not _cupy_available:

            raise Exception('Cupy is not available')
        
        xp = cp if (device == 'gpu' and _cupy_available) else np

        if isinstance(data, (list, tuple, np.ndarray, pd.DataFrame, pd.Series, int, float)):
            data = xp.array(data)
        elif not isinstance(data, (np.ndarray, xp.ndarray if _cupy_available else np.ndarray)):
            # Fallback for unexpected types
            data = xp.array(data)

        self.data = data
        
        self.device = device
        self.requires_grad = requires_grad
        self.grad = xp.zeros_like(self.data, dtype=float) if requires_grad else None
        self._backward = _noop
        global _autograd_enabled
        self._prev = set(_children) if (_autograd_enabled and requires_grad) else set()
        self._op = _op if _autograd_enabled else ''

    def __getstate__(self) -> object:
        return {'data': self.data.get() if self.device == 'gpu' else self.data,
                'device': 'cpu',
                'requires_grad': self.requires_grad}
    
    def __setstate__(self, state):
        self.data = state['data']
        self.device = state['device']
        self.requires_grad = state.get('requires_grad', False)
        xp = cp if (self.device == 'gpu' and _cupy_available) else np
        self.grad = xp.zeros_like(self.data, dtype=float) if self.requires_grad else None
        self._backward = _noop
        self._prev = set()
        self._op = ''

    def __getitem__(self, idx):
        global _autograd_enabled
        out_data = self.data[idx]
        if not _autograd_enabled:
            return Tensor(out_data, device=self.device, requires_grad=self.requires_grad)
        
        out = Tensor(out_data, [self], f'get[{idx}]', device=self.device, requires_grad=self.requires_grad)
        
        def _backward():
            if self.requires_grad:
                xp = cp if self.device == 'gpu' else np
                grad_full = xp.zeros_like(self.data)
                grad_full[idx] = out.grad
                self.grad += grad_full
        
        out._backward = _backward
        return out
    
    def __len__(self):
        return len(self.data)


    def to(self, device):

        if device == self.device:
            return self
        
        if device == "gpu":
            if not _cupy_available:
                raise RuntimeError("CuPy no está instalado, no puedes usar CUDA")
            self.data = cp.asarray(self.data)
            self.grad = cp.array(self.grad) if (self.requires_grad and self.grad is not None) else None
        elif device == "cpu":
            self.data = cp.asnumpy(self.data) if self.device == "gpu" else self.data
            self.grad = cp.asnumpy(self.grad) if (self.requires_grad and self.grad is not None) else None
        else:
            raise ValueError("device debe ser 'cpu' o 'gpu'")
        
        self.device = device

        return self 

    def cpu(self):
        return self.to("cpu")

    def gpu(self):
        return self.to("gpu")
    
    # In-place operations
    def add_(self, other):
        other_data = other.data if isinstance(other, Tensor) else other
        self.data += other_data
        return self

    def sub_(self, other):
        other_data = other.data if isinstance(other, Tensor) else other
        self.data -= other_data
        return self

    def mul_(self, other):
        other_data = other.data if isinstance(other, Tensor) else other
        self.data *= other_data
        return self

    def div_(self, other):
        other_data = other.data if isinstance(other, Tensor) else other
        self.data /= other_data
        return self
        

    def add(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other, device=self.device)
        global _autograd_enabled

        if not _autograd_enabled:
            return Tensor(self.data + other.data, device=self.device)

        requires_grad = self.requires_grad or other.requires_grad   
        out = Tensor(self.data + other.data, [self, other], '+', device=self.device, requires_grad=requires_grad)

        def _backward():
            if out.grad is None:
                return
            if self.requires_grad:
                grad_self = Tensor._match_shape(out.grad, self.data.shape)
                self._accumulate_grad(grad_self)
            if other.requires_grad:
                grad_other = Tensor._match_shape(out.grad, other.data.shape)
                other._accumulate_grad(grad_other)

        out._backward = _backward
        return out

    def __add__(self, other):
        return self.add(other)
    
    def __radd__(self, other):
        return self.add(other)

    def sub(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other, device=self.device)
        global _autograd_enabled

        if not _autograd_enabled:
            return Tensor(self.data - other.data, device=self.device)
        
        requires_grad = self.requires_grad or other.requires_grad
        out = Tensor(self.data - other.data, [self, other], '-', device=self.device, requires_grad=requires_grad)

        def _backward():
            if out.grad is None:
                return
            
            if self.requires_grad:
                grad_self = Tensor._match_shape(out.grad, self.data.shape)
                self._accumulate_grad(grad_self)

            if other.requires_grad:
                grad_other = Tensor._match_shape(out.grad, other.data.shape)
                other._accumulate_grad(-grad_other)

        out._backward = _backward
        return out

    def __sub__(self, other):
        return self.sub(other)
    
    def __rsub__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other, device=self.device)
        return other.sub(self)

    def mul(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other, device=self.device)
        global _autograd_enabled

        if not _autograd_enabled:
            return Tensor(self.data * other.data, device=self.device)
        
        requires_grad = self.requires_grad or other.requires_grad
        out = Tensor(self.data * other.data, [self, other], '*', device=self.device, requires_grad=requires_grad)

        def _backward():
            if out.grad is None:
                return
            
            if self.requires_grad:
                grad_self = Tensor._match_shape(other.data * out.grad, self.data.shape)
                self._accumulate_grad(grad_self)

            if other.requires_grad:
                grad_other = Tensor._match_shape(self.data * out.grad, other.data.shape)
                other._accumulate_grad(grad_other)

        out._backward = _backward
        return out

    def __mul__(self, other):
        return self.mul(other)
    
    def __rmul__(self, other):
        return self.mul(other)

    def matmul(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other, device=self.device)
        global _autograd_enabled

        if not _autograd_enabled :
            return Tensor(self.data @ other.data, device=self.device)
        
        requires_grad = self.requires_grad or other.requires_grad
        out = Tensor(self.data @ other.data, [self, other], '@', device=self.device, requires_grad=requires_grad)

        def _backward():
            if out.grad is None:
                return
            
            if self.requires_grad:
                grad_self = out.grad @ other.data.T
                self._accumulate_grad(Tensor._match_shape(grad_self, self.data.shape))

            if other.requires_grad:
                grad_other = self.data.T @ out.grad
                other._accumulate_grad(Tensor._match_shape(grad_other, other.data.shape))

        out._backward = _backward
        return out

    def tanh(self):
        xp = cp if self.device == 'gpu' else np
        global _autograd_enabled

        if not _autograd_enabled:
            return Tensor(xp.tanh(self.data), device=self.device)
        
        out = Tensor(xp.tanh(self.data), [self], 'tanh', device=self.device, requires_grad=self.requires_grad)

        def _backward():
            if out.grad is None:
                return
            if self.requires_grad:
                self._accumulate_grad(out.grad * (1 - out.data**2))
        
        out._backward = _backward
        return out

    def _accumulate_grad(self, grad):
        if grad is None:
            return
        if self.grad is None:
            xp = cp if self.device == 'gpu' else np
            self.grad = xp.zeros_like(self.data, dtype=float)
        self.grad += grad

    def __matmul__(self, other):
        return self.matmul(other)
    
    def __rmatmul__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other, device=self.device)
        return other.matmul(self)

    def pow(self, other):
        assert isinstance(other, (int, float)), "only supporting int/float powers for now"
        global _autograd_enabled

        if not _autograd_enabled:
            return Tensor(self.data**other, device=self.device)
        
        out = Tensor(self.data**other, [self], f'**{other}', device=self.device, requires_grad=self.requires_grad)

        def _backward():
            if out.grad is None:
                return
            
            if self.requires_grad:
                grad = out.grad * (other * (self.data**(other-1)))
                self._accumulate_grad(grad)

        out._backward = _backward
        return out

    def sigmoid(self):
        xp = cp if self.device == 'gpu' else np
        global _autograd_enabled

        out_data = 1 / (1 + xp.exp(-self.data))
        if not _autograd_enabled:
            return Tensor(out_data, device=self.device, requires_grad=self.requires_grad)
        
        out = Tensor(out_data, [self], 'sigmoid', device=self.device, requires_grad=self.requires_grad)

        def _backward():
            if out.grad is None:
                return
            if self.requires_grad:
                self._accumulate_grad(out.grad * out.data * (1 - out.data))
        
        out._backward = _backward
        return out

    def softmax(self, axis=-1):
        xp = cp if self.device == 'gpu' else np
        global _autograd_enabled
        
        # Stability trick
        shifted_data = self.data - xp.max(self.data, axis=axis, keepdims=True)
        exp_data = xp.exp(shifted_data)
        out_data = exp_data / xp.sum(exp_data, axis=axis, keepdims=True)

        if not _autograd_enabled:
            return Tensor(out_data, device=self.device, requires_grad=self.requires_grad)
        
        out = Tensor(out_data, [self], 'softmax', device=self.device, requires_grad=self.requires_grad)

        def _backward():
            if out.grad is None:
                return
            if self.requires_grad:
                # Softmax gradient: s * (grad - sum(grad * s, axis, keepdims))
                sum_grad_s = xp.sum(out.grad * out.data, axis=axis, keepdims=True)
                self._accumulate_grad(out.data * (out.grad - sum_grad_s))
        
        out._backward = _backward
        return out

    def __pow__(self, other):
        return self.pow(other)

    def div(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other, device=self.device)
        global _autograd_enabled

        if not _autograd_enabled:
            return Tensor(self.data / other.data, device=self.device)
        
        requires_grad = self.requires_grad or other.requires_grad
        out = Tensor(self.data / other.data, [self, other], '/', device=self.device, requires_grad=requires_grad)

        def _backward():
            if out.grad is None:
                return
            
            if self.requires_grad:
                grad_self = Tensor._match_shape(out.grad / other.data, self.data.shape)
                self._accumulate_grad(grad_self)

            if other.requires_grad:
                grad_other = Tensor._match_shape(-self.data * out.grad / (other.data**2), other.data.shape)
                other._accumulate_grad(grad_other)

        out._backward = _backward
        return out

    def __truediv__(self, other):
        return self.div(other)
    
    def __rtruediv__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other, device=self.device)
        return other.div(self)
    
    def mean(self, axis=None, keepdims=False):
        global _autograd_enabled
        xp = cp if self.device == 'gpu' else np

        if not _autograd_enabled:
            return Tensor(xp.mean(self.data, axis=axis, keepdims=keepdims), device=self.device)
        
        out = Tensor(xp.mean(self.data, axis=axis, keepdims=keepdims), [self], 'mean', device=self.device, requires_grad=self.requires_grad)

        def _backward():            
            if out.grad is None:
                return
            
            if self.requires_grad:
                grad = out.grad
                if not keepdims and axis is not None:
                    grad = xp.expand_dims(grad, axis=axis)
                self._accumulate_grad(grad * xp.ones_like(self.data) / self.data.size)
        out._backward = _backward
        return out
    
    def sum(self, axis=None, keepdims=False):
        global _autograd_enabled
        xp = cp if self.device == 'gpu' else np
        
        if not _autograd_enabled:
            return Tensor(self.data.sum(axis=axis, keepdims=keepdims), device=self.device)
            
        out = Tensor(self.data.sum(axis=axis, keepdims=keepdims), [self], 'sum', device=self.device, requires_grad=self.requires_grad)
        
        def _backward():
            if out.grad is None:
                return

            if self.requires_grad:
                grad = out.grad
                if not keepdims and axis is not None:
                    grad = xp.expand_dims(grad, axis=axis)
                self._accumulate_grad(xp.ones_like(self.data) * grad)
        
        out._backward = _backward
        return out
    
    def abs(self):
        xp = cp if self.device == 'gpu' else np
        return Tensor(xp.abs(self.data), device=self.device)
    
    def reshape(self, *shape):
        if len(shape) == 1 and isinstance(shape[0], (list, tuple)):
            shape = shape[0]
            
        global _autograd_enabled
        if not _autograd_enabled:
            return Tensor(self.data.reshape(*shape), device=self.device, requires_grad=self.requires_grad)
        
        out = Tensor(self.data.reshape(*shape), [self], 'reshape', device=self.device, requires_grad=self.requires_grad)
        
        def _backward():
            if out.grad is None:
                return
            if self.requires_grad:
                self._accumulate_grad(out.grad.reshape(self.data.shape))
        
        out._backward = _backward
        return out

    def transpose(self, *axes):
        global _autograd_enabled
        if not _autograd_enabled:
            return Tensor(self.data.transpose(*axes), device=self.device, requires_grad=self.requires_grad)
        
        out = Tensor(self.data.transpose(*axes), [self], 'transpose', device=self.device, requires_grad=self.requires_grad)
        
        def _backward():
            if out.grad is None:
                return
            if self.requires_grad:
                # La inversa de una transposición es la transposición con los ejes invertidos
                if not axes:
                    self._accumulate_grad(out.grad.transpose())
                else:
                    inv_axes = np.argsort(axes)
                    self._accumulate_grad(out.grad.transpose(*inv_axes))
        
        out._backward = _backward
        return out

    @property
    def T(self):
        return self.transpose()

    def flatten(self):
        return self.reshape(-1)

    def backward(self):
        topo = []
        visited = set()

        def build_topo(t):
            if id(t) not in visited:
                visited.add(id(t))
                for child in t._prev:
                    build_topo(child)
                topo.append(t)

        build_topo(self)
        
        xp = cp if self.device == 'gpu' else np
        if self.grad is None:
             self.grad = xp.ones_like(self.data, dtype=float)
        else:
             self.grad += xp.ones_like(self.data, dtype=float)

        for node in reversed(topo):
            node._backward()

    @staticmethod
    def _match_shape(grad, shape):
        if grad is None:
            return None
        while grad.ndim > len(shape):
            grad = grad.sum(axis=0)

        for axis, dim in enumerate(shape):
            if dim == 1 and grad.shape[axis] != 1:
                grad = grad.sum(axis=axis, keepdims=True)
        return grad
    

    def __iter__(self):
        return iter(self.data)

    def __str__(self) -> str:
        return f"Tensor(\n{self.data}, shape={self.data.shape}, device={self.device}, requires_grad={self.requires_grad})"

    def __repr__(self):
        return self.__str__()
    
    @property
    def shape(self):
        return self.data.shape
    
    @property
    def ndim(self):
        return self.data.ndim
    
    @property
    def size(self):
        return self.data.size
    
    @property
    def dtype(self):
        return self.data.dtype
    
    def astype(self, dtype):
        return Tensor(self.data.astype(dtype),device=self.device)
    
    def to_numpy(self):
        return self.data if self.device == 'cpu' else self.data.get()   

    def item(self):
        return self.data.item()
    
    def __array__(self, dtype=None, copy=None):
        arr = self.to_numpy()
        if dtype is not None:
            arr = arr.astype(dtype)
        if copy is False:
             return arr
        return arr.copy()
    
   # Comparaciones
    def __gt__(self, other):
        other_data = other.data if isinstance(other, Tensor) else other
        return Tensor(self.data > other_data, device=self.device, requires_grad=False)

    def __lt__(self, other):
        other_data = other.data if isinstance(other, Tensor) else other
        return Tensor(self.data < other_data, device=self.device, requires_grad=False)

    def __ge__(self, other):
        other_data = other.data if isinstance(other, Tensor) else other
        return Tensor(self.data >= other_data, device=self.device, requires_grad=False)

    def __le__(self, other):
        other_data = other.data if isinstance(other, Tensor) else other
        return Tensor(self.data <= other_data, device=self.device, requires_grad=False)

    def __eq__(self, other):
        other_data = other.data if isinstance(other, Tensor) else other
        return Tensor(self.data == other_data, device=self.device, requires_grad=False)

    def __ne__(self, other):
        other_data = other.data if isinstance(other, Tensor) else other
        return Tensor(self.data != other_data, device=self.device, requires_grad=False)
    
    def __hash__(self):
        return id(self)

def tensor(data, device='cpu', requires_grad=False):
    """
    Factory function to create a Sorix Tensor.
    """
    return Tensor(data, device=device, requires_grad=requires_grad)




from .tensor import Tensor, tensor, no_grad
from .cuda import cuda
from .utils.utils import sigmoid, softmax, argmax
from .utils.utils import (as_tensor, from_numpy,
                          zeros, ones, full, eye, diag, empty,
                          arange, linspace, logspace,
                          rand, randn, randint, randperm,
                          zeros_like, ones_like, empty_like, full_like,
                          save, load
                          )


from .utils.math import (sin, cos, tanh, exp, log, sqrt, mean, sum,
                          add, sub, mul, div, matmul, pow)

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
from typing import Union
from sorix.tensor import Tensor, tensor
from sorix.utils import math as smat
from sorix.utils import utils as sut
from sorix.cupy.cupy import _cupy_available

if _cupy_available:
    import cupy as cp


class Kmeans:

    """
    K-means clustering
    """

    def __init__(self, n_clusters:int):
        """
        Parameters:
        n_clusters (int): number of clusters
        """
        self.n_clusters = n_clusters
        self._centroids = None
        self.features_names = None
        self.labels = None

    def _data_preprocessing_train(self, 
                                 features: Tensor) -> Tensor:

        xp = cp if features.device == 'gpu' and _cupy_available else np

        self._centroids = features.data[xp.random.randint(0, len(features), self.n_clusters)]
        return features

    def _distances(self, features: Tensor, centroids: np.ndarray) -> np.ndarray:
        xp = cp if features.device == 'gpu' and _cupy_available else np

        diff = features.data[:, xp.newaxis, :] - centroids[xp.newaxis, :, :]

        distances = smat.sum(diff**2, axis=-1)

        return distances

    def _new_labels(self, distances: Tensor) -> np.ndarray:

        return sut.argmin(distances, axis=1).flatten()

    def _new_centroids(self, features: Tensor, labels: Tensor) -> np.ndarray:

        xp = cp if features.device == 'gpu' and _cupy_available else np

        k = self.n_clusters

        n_features = features.shape[1]

        # Inicializar acumuladores
        sutms = xp.zeros((k, n_features))
        counts = xp.bincount(labels, minlength=k)

        # Acumular sutmas de cada dimensión con bincount (vectorizado en C)
        for j in range(n_features):
            sutms[:, j] = xp.bincount(labels, weights=features.data[:, j], minlength=k)

        # Evitar división por cero en clusters vacíos
        counts[counts == 0] = 1

        # Calcular centroides = sutma / cantidad
        new_centroids = sutms / counts[:, None]
        return new_centroids 



    def _moviment(self, centroids_before: np.ndarray, centroids_after: np.ndarray) -> float:

        xp = cp if centroids_before.device == 'gpu' and _cupy_available else np

        return xp.linalg.norm(centroids_before - centroids_after, axis=1).mean()


    def fit(self, 
              features: Tensor, 
              eps:float=0.001,
              max_iters:int=1000) -> None:
        
        """
        Fit the model.

        Parameters:
            features (tensor): Features to predict.
            eps (float, optional): Stop criterion, by default 0.001
            max_iters (int, optional): Maximum number of iterations, by default 1000
    
        """
        
        features_train = self._data_preprocessing_train(features)
        iters = 0
        while True:
            iters += 1
            distances = self._distances(features_train, self._centroids)
            self.labels = self._new_labels(distances)
            centroids_before = self._centroids
            self._centroids = self._new_centroids(features_train, self.labels)
            moviment = self._moviment(centroids_before, self._centroids)
            if (moviment < eps) or (iters > max_iters):
                break

    def predict(self, features: Tensor) -> Tensor:

        """
        Predict the labels of features

        Parameters:
            features (tensor): Features to predict.

        Returns:
            labels (tensor): Labels of features
        """

        distances = self._distances(features, self._centroids)
        labels = self._new_labels(distances)
        return tensor(labels)
    
    def get_distances(self, features: Tensor) -> Tensor:

        """
        Get distances between features and centroids

        Parameters:
            features (tensor): Features to predict.

        Returns:
            distances (tensor): Distances betwen features and centroids
        """

        return tensor(self._distances(features, self._centroids))

    def get_inertia(self, features: Tensor) -> float:

        """
        Get inertia of features for k-centroids

        Parameters:
            features (tensor): Features to predict.

        Returns:
            inertia (float): Inertia of features
        """

        distances = self._distances(features, self._centroids)
        labels = self._new_labels(distances)
        
        return smat.sum((features - self.centroids[labels])**2).item()
    
    @property
    def centroids(self) -> Tensor:
        return tensor(self._centroids)

    def __str__(self):
        
        text = f"""
        model: {self.__class__.__name__} \n
        n_centroids: {self.n_clusters} \n
        """
        return text
    
    def __repr__(self):
        return self.__str__()




import pytest
import numpy as np
from sorix import Tensor, tensor
from sorix.nn import Linear, ReLU, Sigmoid, Tanh
from sorix.nn import Module
from sorix.nn import MSELoss, CrossEntropyLoss

def test_linear_layer_forward():
    """Test Linear layer forward pass and parameter management."""
    batch_size = 4
    in_features = 3
    out_features = 2
    
    lin = Linear(in_features, out_features, bias=True)
    x = tensor(np.random.randn(batch_size, in_features))
    
    out = lin(x)
    
    assert out.shape == (batch_size, out_features)
    assert len(lin.parameters()) == 2 # W and b
    assert lin.W.shape == (in_features, out_features)
    assert lin.b.shape == (1, out_features)

def test_linear_layer_backward():
    """Test Linear layer gradients."""
    lin = Linear(2, 1, bias=True)
    # Set weights manually for predictable output
    lin.W.data = np.array([[1.0], [2.0]])
    lin.b.data = np.array([[0.5]])
    
    x = tensor([[1.0, 1.0]], requires_grad=True)
    # out = 1.0*1.0 + 1.0*2.0 + 0.5 = 3.5
    out = lin(x)
    
    out.backward()
    
    # dout/dx = W.T = [[1.0, 2.0]]
    assert np.allclose(x.grad, [[1.0, 2.0]])
    # dout/dW = x.T @ grad_out = [[1.0], [1.0]] @ [[1.0]] = [[1.0], [1.0]]
    assert np.allclose(lin.W.grad, [[1.0], [1.0]])
    # dout/db = sum(grad_out, axis=0) = [[1.0]]
    assert np.allclose(lin.b.grad, [[1.0]])

def test_activations():
    """Test ReLU, Sigmoid, Tanh layers."""
    x_data = np.array([-1.0, 0.0, 1.0])
    x = tensor(x_data, requires_grad=True)
    
    # ReLU
    relu = ReLU()
    out_relu = relu(x)
    assert np.array_equal(out_relu.data, [0.0, 0.0, 1.0])
    out_relu.backward()
    # grad is [0, 0, 1] + 1 if it's root? No, backwards adds 1 to root node's grad.
    # Actually, relu(x) creates a child. out_relu.grad = [1, 1, 1]
    # x.grad = [0, 0, 1]
    assert np.array_equal(x.grad, [0.0, 0.0, 1.0])
    
    x.grad = None
    # Sigmoid
    sig = Sigmoid()
    out_sig = sig(x)
    expected_sig = 1 / (1 + np.exp(-x_data))
    assert np.allclose(out_sig.data, expected_sig)
    out_sig.backward()
    assert np.allclose(x.grad, expected_sig * (1 - expected_sig))

    x.grad = None
    # Tanh
    tanh = Tanh()
    out_tanh = tanh(x)
    expected_tanh = np.tanh(x_data)
    assert np.allclose(out_tanh.data, expected_tanh)
    out_tanh.backward()
    assert np.allclose(x.grad, 1 - expected_tanh**2)

def test_neural_network_parameters():
    """Test parameter collection in Module."""
    class SimpleNet(Module):
        def __init__(self):
            super().__init__()
            self.l1 = Linear(10, 5)
            self.l2 = Linear(5, 2)
            
        def forward(self, x):
            return self.l2(self.l1(x))
            
    net = SimpleNet()
    params = net.parameters()
    
    # Each Linear has 2 params (W, b)
    assert len(params) == 4
    # Check they are actually the same tensors
    assert params[0] is net.l1.W
    assert params[1] is net.l1.b
    assert params[2] is net.l2.W
    assert params[3] is net.l2.b




from sorix.tensor import Tensor, tensor
import numpy as np
from sorix.cupy.cupy import _cupy_available

if _cupy_available:
    import cupy as cp
    

from .net import Module

class Linear(Module):
    def __init__(self, features: int, neurons: int,bias=True, init='he',device='cpu'):
        super().__init__()
        if device == 'gpu' and not _cupy_available:
            raise Exception('Cupy is not available')
        
        self.device = device

        xp = cp if device == 'gpu' else np
        
        if init not in ['he', 'xavier']:
            raise ValueError(f'Invalid initialization method: {init}. Valid methods are "he" and "xavier"')
        
        if init == 'he':
            self.std_dev = xp.sqrt(2.0 / features)  # He init para ReLU
        elif init == 'xavier':
            self.std_dev = xp.sqrt(2.0 / (features + neurons))  # Xavier init para tanh

        self.bias = bias
        self.W = tensor(xp.random.normal(0, self.std_dev, size=(features, neurons)),device=self.device,requires_grad=True)
        self.b = tensor(xp.zeros((1, neurons)),device=self.device,requires_grad=True)  if self.bias else None

    def __call__(self, X: Tensor):
        if self.bias:
            return X @ self.W + self.b  
        return X @ self.W
    
    @property
    def coef_(self):
        return self.W.data.flatten()
        
    @property
    def intercept_(self):
        return self.b.item() if self.b is not None else None


class ReLU(Module):
    def __call__(self, X: Tensor):

        xp = cp if X.device == 'gpu' else np

        out = Tensor(xp.maximum(0, X.data), (X,), 'ReLU',device=X.device,requires_grad=X.requires_grad)
        
        def _backward():
            if out.grad is None:
                return
            if X.requires_grad:
                X._accumulate_grad(out.grad * (X.data > 0))
        out._backward = _backward
        return out


class Sigmoid(Module):
    def __call__(self, X: Tensor):

        xp = cp if X.device == 'gpu' else np

        out = Tensor(1 / (1 + xp.exp(-X.data)), (X,), 'Sigmoid',device=X.device,requires_grad=X.requires_grad)
        
        def _backward():
            if out.grad is None:
                return
            if X.requires_grad:
                X._accumulate_grad(out.grad * out.data * (1 - out.data))
        out._backward = _backward
        return out
    
    
class Tanh(Module):
    def __call__(self, X: Tensor):

        xp = cp if X.device == 'gpu' else np

        out = Tensor(xp.tanh(X.data), (X,), 'Tanh',device=X.device,requires_grad=X.requires_grad)
        
        def _backward():
            if out.grad is None:
                return
            if X.requires_grad:
                X._accumulate_grad(out.grad * (1 - out.data**2))
        out._backward = _backward
        return out

class BatchNorm1d(Module):
    def __init__(self, features: int, alpha: float = 0.9, epsilon: float = 1e-5, device='cpu'):
        super().__init__()
        self.gamma = tensor(np.ones((1, features)), requires_grad=True)
        self.beta = tensor(np.zeros((1, features)), requires_grad=True)

        # buffers (no requieren gradiente, ahora son tensores para que state_dict los capture)
        self.running_mean = tensor(np.zeros((1, features), dtype=np.float32), requires_grad=False)
        self.running_var = tensor(np.ones((1, features), dtype=np.float32), requires_grad=False)

        self.alpha = alpha
        self.epsilon = epsilon
        self.device = device
        self.training = True
        
        if self.device != 'cpu':
            self.to(self.device)

    def __call__(self, X: Tensor):
        xp = cp if X.device == 'gpu' else np

        if self.training:
            # estadísticas del batch
            batch_mean = xp.mean(X.data, axis=0, keepdims=True)
            batch_var = xp.var(X.data, axis=0, keepdims=True)

            # actualizar los buffers (actualizamos .data directamente para eficiencia)
            self.running_mean.data = self.alpha * self.running_mean.data + (1 - self.alpha) * batch_mean
            self.running_var.data = self.alpha * self.running_var.data + (1 - self.alpha) * batch_var

            mean = batch_mean
            var = batch_var
        else:
            # usar estadísticas acumuladas
            mean = self.running_mean.data
            var = self.running_var.data

        # normalizar
        X_norm = (X - mean) / xp.sqrt(var + self.epsilon)
        out = self.gamma * X_norm + self.beta
        return out



from sorix.tensor import Tensor, tensor
import numpy as np
from sorix.cupy.cupy import _cupy_available

if _cupy_available:
    try:
        import cupy as cp
    except ImportError:
        _cupy_available = False
        cp = None
else:
    cp = None


def sigmoid(X) -> Tensor | np.ndarray:
    if isinstance(X, Tensor):
        return X.sigmoid()
    
    xp = cp if (_cupy_available and (cp is not None and isinstance(X, cp.ndarray))) else np
    return 1 / (1 + xp.exp(-X))


def softmax(X, axis=-1) -> Tensor | np.ndarray:
    if isinstance(X, Tensor):
        return X.softmax(axis=axis)
    
    xp = cp if (_cupy_available and (cp is not None and isinstance(X, cp.ndarray))) else np
    exp_logits = xp.exp(X - xp.max(X, axis=axis, keepdims=True))
    return exp_logits / xp.sum(exp_logits, axis=axis, keepdims=True)


def argmax(X, axis=1, keepdims=True) -> Tensor | np.ndarray:

    if isinstance(X, Tensor):
        xp = cp if X.device == 'gpu' and _cupy_available else np
        return tensor(xp.argmax(X.data, axis=axis, keepdims=keepdims),device=X.device)
    else:
        xp = cp if (cp is not None and isinstance(X, cp.ndarray)) else np
        return X.argmax(axis=axis, keepdims=keepdims)
    
def argmin(X, axis=1, keepdims=True) -> Tensor | np.ndarray:

    if isinstance(X, Tensor):
        xp = cp if X.device == 'gpu' and _cupy_available else np
        return tensor(xp.argmin(X.data, axis=axis, keepdims=keepdims),device=X.device)
    else:
        xp = cp if (cp is not None and isinstance(X, cp.ndarray)) else np
        return X.argmin(axis=axis, keepdims=keepdims)
    

def as_tensor(x):
    if isinstance(x, Tensor):
        return x

    return tensor(x)

def from_numpy(x):
    if isinstance(x, Tensor):
        return x

    return tensor(x)

def zeros(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.zeros(*args),device=device,requires_grad=requires_grad)


def ones(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.ones(*args),device=device,requires_grad=requires_grad) 


def full(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.full(*args),device=device,requires_grad=requires_grad)


def eye(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.eye(*args),device=device,requires_grad=requires_grad)

def diag(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.diag(*args),device=device,requires_grad=requires_grad)


def empty(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.empty(*args),device=device,requires_grad=requires_grad)

def arange(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.arange(*args),device=device,requires_grad=requires_grad)

def linspace(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.linspace(*args),device=device,requires_grad=requires_grad)

def logspace(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.logspace(*args),device=device,requires_grad=requires_grad)




def rand(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.random.rand(*args),device=device,requires_grad=requires_grad)


def randn(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.random.randn(*args),device=device,requires_grad=requires_grad)

def randint(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.random.randint(*args),device=device,requires_grad=requires_grad)


def randperm(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.random.permutation(*args),device=device,requires_grad=requires_grad)


def zeros_like(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.zeros_like(*args),device=device,requires_grad=requires_grad)


def ones_like(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.ones_like(*args),device=device,requires_grad=requires_grad)

def empty_like(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.empty_like(*args),device=device,requires_grad=requires_grad)


def full_like(*args,device='cpu',requires_grad=False):

    if device == 'gpu' and not _cupy_available:
        raise Exception('Cupy is not available')
    
    xp = cp if device == 'gpu' and _cupy_available else np

    return tensor(xp.full_like(*args),device=device,requires_grad=requires_grad)

import pickle

def save(obj, f):
    """
    Saves an object to a file using pickle. 
    Tensors will be automatically moved to CPU during serialization.
    """
    if isinstance(f, str):
        with open(f, 'wb') as file:
            pickle.dump(obj, file)
    else:
        pickle.dump(obj, f)

def load(f):
    """
    Loads an object from a file using pickle.
    """
    if isinstance(f, str):
        with open(f, 'rb') as file:
            return pickle.load(file)
    else:
        return pickle.load(f)




import pytest
import numpy as np
from sorix import Tensor, tensor
from sorix.metrics import (
    mean_squared_error, 
    root_mean_squared_error, 
    mean_absolute_error, 
    mean_absolute_percentage_error, 
    r2_score,
    accuracy_score,
    confusion_matrix,
    classification_report,
    regression_report
)

def test_regression_metrics():
    y_true = tensor([1.0, 2.0, 3.0])
    y_pred = tensor([1.1, 1.9, 3.2])
    
    # MSE: ((0.1)^2 + (-0.1)^2 + (0.2)^2) / 3 = (0.01 + 0.01 + 0.04) / 3 = 0.06 / 3 = 0.02
    assert np.isclose(mean_squared_error(y_true, y_pred), 0.02)
    assert np.isclose(root_mean_squared_error(y_true, y_pred), np.sqrt(0.02))
    
    # MAE: (0.1 + 0.1 + 0.2) / 3 = 0.4 / 3 = 0.1333
    assert np.isclose(mean_absolute_error(y_true, y_pred), 0.4 / 3.0)
    
    # r2_score
    # sy = Var(y_true) * n = ((1-2)^2 + (2-2)^2 + (3-2)^2) / 3 * 3 = 2.0 / 3 * 3 = 2.0
    # sr = MSE * 3 = 0.06
    # R2 = 1 - (0.06 / 2.0) = 1 - 0.03 = 0.97
    # Wait, sorix r2_score uses mean() for both, so it's 1 - (MSE / Var)
    # R2 = 1 - (0.02 / (2.0/3)) = 1 - 0.03 = 0.97
    assert np.isclose(r2_score(y_true, y_pred), 0.97)

def test_accuracy_score():
    y_true = tensor([0, 1, 1, 0])
    y_pred = tensor([0, 1, 0, 0])
    assert accuracy_score(y_true, y_pred) == 0.75

def test_confusion_matrix():
    y_true = tensor([0, 1, 2, 2, 0])
    y_pred = tensor([0, 2, 2, 2, 1])
    cm = confusion_matrix(y_true, y_pred)
    
    # Classes: 0, 1, 2
    # y_true 0: pred 0 (1), pred 1 (1) -> row 0: [1, 1, 0]
    # y_true 1: pred 2 (1)           -> row 1: [0, 0, 1]
    # y_true 2: pred 2 (2)           -> row 2: [0, 0, 2]
    expected = np.array([
        [1, 1, 0],
        [0, 0, 1],
        [0, 0, 2]
    ])
    assert np.array_equal(cm, expected)

def test_reports_run_without_error():
    y_true = np.array([0, 1, 0, 1])
    y_pred = np.array([0, 1, 1, 1])
    
    # Just check if they return strings and don't crash
    reg_rep = regression_report(tensor(y_true), tensor(y_pred))
    class_rep = classification_report(tensor(y_true), tensor(y_pred))
    
    assert isinstance(reg_rep, str)
    assert isinstance(class_rep, str)
    assert "MAE" in reg_rep
    assert "precision" in class_rep



from sorix.tensor import Tensor, tensor

class Module:
    def __init__(self):
        super().__init__()
        self.device = 'cpu'
        self.training = True

    def forward(self, x: Tensor) -> Tensor:
        raise NotImplementedError("Debes implementar forward en la subclase.")

    def parameters(self):
        """
        Returns an iterator over module parameters.
        """
        params = []
        visited = set()

        def _gather_params(obj):
            if id(obj) in visited:
                return
            visited.add(id(obj))

            if isinstance(obj, Tensor):
                if obj.requires_grad:
                    params.append(obj)
            elif hasattr(obj, "parameters") and callable(obj.parameters) and obj is not self:
                # If the object has its own parameters() method, use it
                params.extend(obj.parameters())
            elif hasattr(obj, "__dict__"):
                # Recurse into attributes
                for k, v in obj.__dict__.items():
                    if not k.startswith('_'):
                        _gather_params(v)
            elif isinstance(obj, (list, tuple)):
                for item in obj:
                    _gather_params(item)
            elif isinstance(obj, dict):
                for item in obj.values():
                    _gather_params(item)

        _gather_params(self)
        return params

    def to(self, device):
        """Mueve TODOS los tensores/capas/subredes al device."""
        self.device = device

        def _apply(obj):
            if hasattr(obj, "to") and callable(obj.to) and obj is not self:
                return obj.to(device)
            
            if isinstance(obj, Tensor):
                return obj.to(device)

            if isinstance(obj, list):
                return [_apply(v) for v in obj]
            if isinstance(obj, tuple):
                return tuple(_apply(v) for v in obj)
            if isinstance(obj, dict):
                return {k: _apply(v) for k, v in obj.items()}
            
            if hasattr(obj, "__dict__") and obj is not self:
                 # Try to move attributes of non-Module objects
                 for k, v in obj.__dict__.items():
                     if not k.startswith('_'):
                         setattr(obj, k, _apply(v))
            
            return obj

        for k, v in self.__dict__.items():
            if not k.startswith('_'):
                setattr(self, k, _apply(v))

        return self

    def train(self):
        self.training = True
        def _apply(obj):
            if hasattr(obj, "training"):
                obj.training = True
            if isinstance(obj, (list, tuple)):
                for o in obj: _apply(o)
            elif isinstance(obj, dict):
                for v in obj.values(): _apply(v)
        for k, v in self.__dict__.items():
            if not k.startswith('_'):
                _apply(v)

    def eval(self):
        self.training = False
        def _apply(obj):
            if hasattr(obj, "training"):
                obj.training = False
            if isinstance(obj, (list, tuple)):
                for o in obj: _apply(o)
            elif isinstance(obj, dict):
                for v in obj.values(): _apply(v)
        for k, v in self.__dict__.items():
            if not k.startswith('_'):
                _apply(v)

    def __call__(self, x: Tensor) -> Tensor:
        return self.forward(x)
    
    def state_dict(self):
        """
        Returns a dictionary containing a whole state of the module.
        """
        state = {}

        def _get_state(obj, prefix):
            for name, val in obj.__dict__.items():
                if name.startswith('_') or name == 'device' or name == 'training':
                    continue
                
                key = prefix + name
                if isinstance(val, Tensor):
                    state[key] = val
                elif hasattr(val, 'state_dict') and callable(val.state_dict) and val is not self:
                    sub_state = val.state_dict()
                    for sk, sv in sub_state.items():
                        state[key + '.' + sk] = sv
                elif hasattr(val, '__dict__'):
                     for kn, kv in val.__dict__.items():
                         if isinstance(kv, Tensor):
                             state[key + '.' + kn] = kv
            
        _get_state(self, "")
        return state

    def load_state_dict(self, state_dict):
        """
        Copies parameters and buffers from state_dict into this module and its descendants.
        """
        own_state = self.state_dict()
        for name, param in state_dict.items():
            if name in own_state:
                if isinstance(param, Tensor):
                    own_state[name].data = param.data
                    own_state[name].to(own_state[name].device) # Ensure device consistency
                else:
                    pass
            else:
                pass

    def weights(self):
        """Deprecated: use state_dict() instead."""
        return self.state_dict()

    def load_weights(self, weights):
        """Deprecated: use load_state_dict() instead."""
        self.load_state_dict(weights)





from sorix.tensor import Tensor, tensor
import numpy as np
from sorix.cupy.cupy import _cupy_available

if _cupy_available:
    try:
        import cupy as cp
    except ImportError:
        _cupy_available = False
        cp = None
else:
    cp = None


def add(input, other):
    if isinstance(input, Tensor):
        return input.add(other)
    return input + other

def sub(input, other):
    if isinstance(input, Tensor):
        return input.sub(other)
    return input - other

def mul(input, other):
    if isinstance(input, Tensor):
        return input.mul(other)
    return input * other

def div(input, other):
    if isinstance(input, Tensor):
        return input.div(other)
    return input / other

def matmul(input, other):
    if isinstance(input, Tensor):
        return input.matmul(other)
    return input @ other

def pow(input, exponent):
    if isinstance(input, Tensor):
        return input.pow(exponent)
    return input ** exponent

def sin(X):
    if isinstance(X, Tensor):
        xp = cp if X.device == 'gpu' and _cupy_available else np
        out = Tensor(xp.sin(X.data), (X,), 'sin', device=X.device, requires_grad=X.requires_grad)

        def _backward():
            if out.grad is None:
                return
            if X.requires_grad:
                X.grad += out.grad * xp.cos(X.data)  # d/dx sin(x) = cos(x)

        out._backward = _backward
        return out
    else:
        xp = cp if (cp is not None and isinstance(X, cp.ndarray)) else np
        return xp.sin(X)


def cos(X):
    if isinstance(X, Tensor):
        xp = cp if X.device == 'gpu' and _cupy_available else np
        out = Tensor(xp.cos(X.data), (X,), 'cos', device=X.device, requires_grad=X.requires_grad)

        def _backward():
            if out.grad is None:
                return
            if X.requires_grad:
                X.grad += -out.grad * xp.sin(X.data)  # d/dx cos(x) = -sin(x)

        out._backward = _backward
        return out
    else:
        xp = cp if (cp is not None and isinstance(X, cp.ndarray)) else np
        return xp.cos(X)
    
def tanh(X):
    if isinstance(X, Tensor):
        return X.tanh()
    else:
        xp = cp if (cp is not None and isinstance(X, cp.ndarray)) else np
        return xp.tanh(X)
    

def exp(X):
    if isinstance(X, Tensor):
        xp = cp if X.device == 'gpu' and _cupy_available else np
        out = Tensor(xp.exp(X.data), (X,), 'exp', device=X.device, requires_grad=X.requires_grad)

        def _backward():
            if out.grad is None:
                return
            if X.requires_grad:
                X.grad += out.grad * out.data

        out._backward = _backward
        return out
    else:
        xp = cp if (cp is not None and isinstance(X, cp.ndarray)) else np
        return xp.exp(X)
    

def log(X):
    if isinstance(X, Tensor):
        xp = cp if X.device == 'gpu' and _cupy_available else np
        out = Tensor(xp.log(X.data), (X,), 'log', device=X.device, requires_grad=X.requires_grad)

        def _backward():
            if out.grad is None:
                return
            if X.requires_grad:
                X.grad += out.grad / X.data

        out._backward = _backward
        return out
    else:
        xp = cp if (cp is not None and isinstance(X, cp.ndarray)) else np
        return xp.log(X)
    

def sqrt(X):
    if isinstance(X, Tensor):
        xp = cp if X.device == 'gpu' and _cupy_available else np
        out = Tensor(xp.sqrt(X.data), (X,), 'sqrt', device=X.device, requires_grad=X.requires_grad)


        def _backward():
            if out.grad is None:
                return
            if X.requires_grad:
                X.grad += out.grad / (2 * out.data)

        out._backward = _backward
        return out
    else:
        xp = cp if (cp is not None and isinstance(X, cp.ndarray)) else np
        return xp.sqrt(X)
    
def mean(X, axis=None, keepdims=False):
    if isinstance(X, Tensor):
        return X.mean(axis=axis, keepdims=keepdims)
    else:
        xp = cp if (cp is not None and isinstance(X, cp.ndarray)) else np
        return xp.mean(X, axis=axis, keepdims=keepdims)
    

def sum(X, axis=None, keepdims=False):
    if isinstance(X, Tensor):
        return X.sum(axis=axis, keepdims=keepdims)
    else:
        xp = cp if (cp is not None and isinstance(X, cp.ndarray)) else np
        return xp.sum(X, axis=axis, keepdims=keepdims)



from pathlib import Path
import mkdocs_gen_files

nav = mkdocs_gen_files.Nav()

for path in sorted(Path("sorix").rglob("*.py")):
    module_path = path.with_suffix("")
    doc_path = path.relative_to("sorix").with_suffix(".md")
    full_doc_path = Path("api", doc_path)

    parts = list(module_path.parts)

    if parts[-1] == "__init__":
        parts.pop()
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
    elif parts[-1] == "__main__":
        continue

    # Clean navigation parts (remove root package name if it's 'sorix')
    nav_parts = list(parts)
    if nav_parts[0] == "sorix":
        nav_parts.pop(0)
    
    # If it's the root __init__, it becomes empty, so we name it after the package
    if not nav_parts:
        nav_parts = ["sorix"]

    nav[tuple(nav_parts)] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        ident = ".".join(parts)
        fd.write(f"::: {ident}")

    mkdocs_gen_files.set_edit_path(full_doc_path, path)

generated_nav = nav.build_literate_nav()
with mkdocs_gen_files.open("api/SUMMARY.md", "w") as nav_file:
    nav_file.writelines(generated_nav)



import pytest
import numpy as np
import os
import joblib
from sorix import Tensor, tensor, no_grad
from sorix.nn import Linear, ReLU
from sorix import nn
from sorix.nn import MSELoss
from sorix.optim import Adam
from sorix.cuda import cuda

class SimpleNet(nn.Module):
    def __init__(self, in_features, out_features):
        super().__init__()
        self.l1 = Linear(in_features, 10)
        self.relu = ReLU()
        self.l2 = Linear(10, out_features)
        
    def forward(self, x):
        x = self.l1(x)
        x = self.relu(x)
        x = self.l2(x)
        return x

def train_model(device='cpu'):
    # Generate synthetic data for y = x1 + x2
    X_raw = np.random.randn(200, 2).astype(np.float32)
    y_raw = (X_raw[:, 0] + X_raw[:, 1]).reshape(-1, 1)
    
    X = tensor(X_raw, device=device)
    y = tensor(y_raw, device=device)
    
    model = SimpleNet(2, 1).to(device)
    optimizer = Adam(model.parameters(), lr=0.01)
    criterion = MSELoss()
    
    initial_loss = criterion(model(X), y).item()
    
    # Fast training loop
    for epoch in range(150):
        optimizer.zero_grad()
        pred = model(X)
        loss = criterion(pred, y)
        loss.backward()
        optimizer.step()
    
    final_loss = loss.item()
    return model, initial_loss, final_loss

def test_training_cpu_convergence():
    """Verify that a model can converge on CPU."""
    model, initial_loss, final_loss = train_model(device='cpu')
    
    assert final_loss < initial_loss
    assert final_loss < 0.1
    print(f"CPU Training: Initial Loss {initial_loss:.4f}, Final Loss {final_loss:.4f}")

@pytest.mark.skipif(not cuda.is_available(verbose=False), reason="CUDA not available")
def test_training_gpu_convergence():
    """Verify that a model can converge on GPU if available."""
    model, initial_loss, final_loss = train_model(device='gpu')
    
    assert final_loss < initial_loss
    assert final_loss < 0.1
    print(f"GPU Training: Initial Loss {initial_loss:.4f}, Final Loss {final_loss:.4f}")

def test_model_save_and_load(tmp_path):
    """Test exporting and importing model weights using the new API."""
    import sorix
    model, _, _ = train_model(device='cpu')
    
    # Test data
    test_x_raw = np.array([[1.0, 2.0]], dtype=np.float32)
    test_x = tensor(test_x_raw)
    
    with no_grad():
        original_output = model(test_x).to_numpy()
    
    # Save weights using state_dict and sorix.save
    weight_path = os.path.join(tmp_path, "model.sor")
    sorix.save(model.state_dict(), weight_path)
    
    # Create new model and load weights using state_dict and sorix.load
    new_model = SimpleNet(2, 1)
    loaded_state = sorix.load(weight_path)
    new_model.load_state_dict(loaded_state)
    
    with no_grad():
        new_output = new_model(test_x).to_numpy()
    
    # Outputs should be identical
    assert np.allclose(original_output, new_output, atol=1e-6)
    print("Model serialization test (state_dict API) passed.")


def test_to_device_consistency():
    """Verify that moving a model between devices preserves weights."""
    model = SimpleNet(5, 2)
    W_before = model.l1.W.to_numpy().copy()
    
    # To CPU (should be same)
    model.to('cpu')
    assert np.allclose(model.l1.W.to_numpy(), W_before)
    
    # If GPU is available, test round-trip
    if cuda.is_available(verbose=False):
        model.to('gpu')
        assert model.l1.W.device == 'gpu'
        model.to('cpu')
        assert model.l1.W.device == 'cpu'
        assert np.allclose(model.l1.W.to_numpy(), W_before)



import numpy as np
from sorix.tensor import Tensor, tensor


def _get_metric_data(Y_true, Y_pred):
    if isinstance(Y_true, Tensor):
        return Y_true, Y_pred
    return np.array(Y_true), np.array(Y_pred)

def mean_squared_error(Y_true, Y_pred):
    Y_true, Y_pred = _get_metric_data(Y_true, Y_pred)
    mse = ((Y_true-Y_pred)**2).mean()
    return mse.item() if hasattr(mse, 'item') else float(mse)

def root_mean_squared_error(Y_true, Y_pred):
    return mean_squared_error(Y_true, Y_pred)**0.5


def mean_absolute_error(Y_true, Y_pred):
    Y_true, Y_pred = _get_metric_data(Y_true, Y_pred)
    if isinstance(Y_true, Tensor):
        mae = (Y_true-Y_pred).abs().mean()
    else:
        mae = np.abs(Y_true-Y_pred).mean()
    return mae.item() if hasattr(mae, 'item') else float(mae)

def mean_absolute_percentage_error(Y_true, Y_pred):
    Y_true, Y_pred = _get_metric_data(Y_true, Y_pred)
    if isinstance(Y_true, Tensor):
        mape = ((Y_true-Y_pred)/Y_true).abs().mean()
    else:
        mape = np.abs((Y_true-Y_pred)/Y_true).mean()
    return mape.item() if hasattr(mape, 'item') else float(mape)

def r2_score(Y_true, Y_pred):
    Y_true, Y_pred = _get_metric_data(Y_true, Y_pred)
    sr = ((Y_true-Y_pred)**2).mean()
    sy = ((Y_true-Y_true.mean())**2).mean()
    r2 = (1-(sr/sy))
    return r2.item() if hasattr(r2, 'item') else float(r2)



def regression_report(y_true: np.ndarray, y_pred: np.ndarray) -> str:
    """
    Reporte de regresión con columnas alineadas y rango uniforme.
    """
    metrics = {
        "R2":   (r2_score(y_true, y_pred), "[0,   1]"),
        "MAE":  (mean_absolute_error(y_true, y_pred), "[0,  ∞)"),
        "MSE":  (mean_squared_error(y_true, y_pred), "[0,  ∞)"),
        "RMSE": (root_mean_squared_error(y_true, y_pred), "[0,  ∞)"),
        "MAPE": (mean_absolute_percentage_error(y_true, y_pred) * 100, "[0, 100]"),
    }

    # Forzar todos los rangos a la misma longitud (8 caracteres)
    fixed_width = 8
    for k, (val, rng) in metrics.items():
        metrics[k] = (val, rng.ljust(fixed_width))

    col_metric = 6
    col_score = 9
    col_range = fixed_width

    header = f"{'Metric':<{col_metric}} | {'Score':>{col_score}} | {'Range':>{col_range}}"
    lines = [header, "-" * len(header)]

    for name, (value, rng) in metrics.items():
        lines.append(f"{name:<{col_metric}} | {value:>{col_score}.4f} | {rng:>{col_range}}")

    return "\n".join(lines)


def accuracy_score(Y_true, Y_pred):
    Y_true, Y_pred = _get_metric_data(Y_true, Y_pred)
    if isinstance(Y_true, Tensor):
        acc = (Y_true == Y_pred).mean()
    else:
        acc = (Y_true == Y_pred).mean()
    return acc.item() if hasattr(acc, 'item') else float(acc)


def confusion_matrix(y_true, y_pred):
    y_true, y_pred = _get_classification_data(y_true, y_pred)
    
    classes = np.unique(y_true)
    cm = np.zeros((len(classes), len(classes)))

    for i, c1 in enumerate(classes):
        for j, c2 in enumerate(classes):
            cm[i, j] = np.sum((y_true == c1) & (y_pred == c2))
    cm = cm.astype(int)

    return cm




def _get_classification_data(y_true, y_pred):
    if y_true.shape != y_pred.shape:
        raise ValueError("y_true and y_pred must have the same shape.")
    
    if isinstance(y_true, Tensor):
        y_true = y_true.to_numpy().flatten()
    elif isinstance(y_true, (list, tuple)):
        y_true = np.array(y_true).flatten()

    if isinstance(y_pred, Tensor):
        y_pred = y_pred.to_numpy().flatten()
    elif isinstance(y_pred, (list, tuple)):
        y_pred = np.array(y_pred).flatten()
        
    return y_true, y_pred

def precision_score(y_true, y_pred, average='binary', pos_label=1):
    y_true, y_pred = _get_classification_data(y_true, y_pred)
    classes = np.unique(y_true)
    
    if average == 'binary':
        true_pos = np.sum((y_true == pos_label) & (y_pred == pos_label))
        pred_pos = np.sum(y_pred == pos_label)
        return true_pos / pred_pos if pred_pos > 0 else 0.0
    
    precisions = []
    supports = []
    for c in classes:
        true_pos = np.sum((y_true == c) & (y_pred == c))
        pred_pos = np.sum(y_pred == c)
        precisions.append(true_pos / pred_pos if pred_pos > 0 else 0.0)
        supports.append(np.sum(y_true == c))
        
    if average == 'macro':
        return np.mean(precisions)
    elif average == 'weighted':
        return np.average(precisions, weights=supports)
    return precisions

def recall_score(y_true, y_pred, average='binary', pos_label=1):
    y_true, y_pred = _get_classification_data(y_true, y_pred)
    classes = np.unique(y_true)
    
    if average == 'binary':
        true_pos = np.sum((y_true == pos_label) & (y_pred == pos_label))
        actual_pos = np.sum(y_true == pos_label)
        return true_pos / actual_pos if actual_pos > 0 else 0.0
    
    recalls = []
    supports = []
    for c in classes:
        true_pos = np.sum((y_true == c) & (y_pred == c))
        actual_pos = np.sum(y_true == c)
        recalls.append(true_pos / actual_pos if actual_pos > 0 else 0.0)
        supports.append(actual_pos)
        
    if average == 'macro':
        return np.mean(recalls)
    elif average == 'weighted':
        return np.average(recalls, weights=supports)
    return recalls

def f1_score(y_true, y_pred, average='binary', pos_label=1):
    p = precision_score(y_true, y_pred, average=average, pos_label=pos_label)
    r = recall_score(y_true, y_pred, average=average, pos_label=pos_label)
    
    if isinstance(p, (list, np.ndarray)):
        p = np.array(p)
        r = np.array(r)
        denom = p + r
        denom[denom == 0] = 1e-9
        return 2 * p * r / denom
    
    return 2 * p * r / (p + r) if (p + r) > 0 else 0.0

def classification_report(y_true: np.ndarray, y_pred: np.ndarray) -> str:
    """
    Reporte de clasificación similar a sklearn.metrics.classification_report.
    """
    y_true, y_pred = _get_classification_data(y_true, y_pred)
    classes = sorted(np.unique(y_true))
    report = {}

    total_true = len(y_true)

    # Métricas por clase
    for c in classes:
        true_pos = np.sum((y_true == c) & (y_pred == c))
        pred_pos = np.sum(y_pred == c)
        actual_pos = np.sum(y_true == c)

        precision = true_pos / pred_pos if pred_pos > 0 else 0.0
        recall = true_pos / actual_pos if actual_pos > 0 else 0.0
        f1 = (2 * precision * recall / (precision + recall)
              if (precision + recall) > 0 else 0.0)

        report[c] = {
            "precision": precision,
            "recall": recall,
            "f1": f1,
            "support": actual_pos
        }

    # Promedio macro
    macro_precision = np.mean([report[c]["precision"] for c in classes])
    macro_recall = np.mean([report[c]["recall"] for c in classes])
    macro_f1 = np.mean([report[c]["f1"] for c in classes])

    # Promedio ponderado
    weights = np.array([report[c]["support"] for c in classes])
    weighted_precision = np.average([report[c]["precision"] for c in classes], weights=weights)
    weighted_recall = np.average([report[c]["recall"] for c in classes], weights=weights)
    weighted_f1 = np.average([report[c]["f1"] for c in classes], weights=weights)

    # Estilo sklearn: ancho fijo y espacio inicial
    header = f"{'':<12}{'precision':>9}{'recall':>9}{'f1-score':>9}{'support':>9}"
    lines = [header]
    
    # Líneas para cada clase
    for c in classes:
        lines.append(f"{str(c):<12}{report[c]['precision']:>9.2f}{report[c]['recall']:>9.2f}{report[c]['f1']:>9.2f}{report[c]['support']:>9}")
    
    lines.append("")
    
    # Fila de accuracy (sklearn también la incluye)
    accuracy = np.sum(y_true == y_pred) / total_true
    lines.append(f"{'accuracy':<12}{'':>9}{'':>9}{accuracy:>9.2f}{total_true:>9}")
    
    # Líneas de promedios con alineación correcta
    lines.append(f"{'macro avg':<12}{macro_precision:>9.2f}{macro_recall:>9.2f}{macro_f1:>9.2f}{total_true:>9}")
    lines.append(f"{'weighted avg':<12}{weighted_precision:>9.2f}{weighted_recall:>9.2f}{weighted_f1:>9.2f}{total_true:>9}")

    return "\n".join(lines)



